# Copyright 2022 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################

from mio import user
from mio import common
from mio import cache
from mio import cfg
from mio import sim
from jinja2 import Template
from fusesoc import main as fsoc
from alive_progress import alive_bar
import fusesoc
import atexit
import pathlib
import argparse
import os
import subprocess
import re
import yaml
import glob
from yaml.loader import SafeLoader
from threading import BoundedSemaphore

eda_processes = []
bar = None

vivado_default_compilation_args  = ["--incr", "-sv"]
metrics_default_compilation_args = ["+acc+b", "-suppress MultiBlockWrite:ReadingOutputModport:UndefinedMacro"]
vcs_default_compilation_args     = ["-lca", "-sverilog"]
xcelium_default_compilation_args = []
questa_default_compilation_args  = ["-64", "-incrcomp"]
riviera_default_compilation_args = []

vivado_project_default_vlog_compilation_args = ["--relax"]
vivado_project_default_vhdl_compilation_args = ["--relax"]

vivado_default_elaboration_args  = ["--incr", "-relax", "--O0", "-v 0", "-timescale 1ns/1ps", "-dup_entity_as_module"]
metrics_default_elaboration_args = ["+acc+b"]
vcs_default_elaboration_args     = []
xcelium_default_elaboration_args = []
questa_default_elaboration_args  = ["-64"]
riviera_default_elaboration_args = []

vivado_default_simulation_args  = ["--stats"]
metrics_default_simulation_args = []
vcs_default_simulation_args     = []
xcelium_default_simulation_args = []
questa_default_simulation_args  = ["-64", "-c"]
riviera_default_simulation_args = []

vivado_cmp_log_error_regexes  = ["ERROR:", "CRITICAL WARNING:"]
metrics_cmp_log_error_regexes = ["=E:", "=F:"]
vcs_cmp_log_error_regexes     = ["Error-"]
xcelium_cmp_log_error_regexes = ["*E "]
questa_cmp_log_error_regexes  = ["\*\* Error:"]
riviera_cmp_log_error_regexes = ["Error:"]

vivado_cmp_log_warning_regexes  = ["WARNING:"]
metrics_cmp_log_warning_regexes = ["=W:"]
vcs_cmp_log_warning_regexes     = ["Warning-"]
xcelium_cmp_log_warning_regexes = ["*W "]
questa_cmp_log_warning_regexes  = ["\*\* Warning:"]
riviera_cmp_log_warning_regexes = ["Warning:"]

vivado_elab_log_error_regexes  = ["ERROR:","Invalid path for DPI library:"]
metrics_elab_log_error_regexes = ["=E:", "=F:"]
vcs_elab_log_error_regexes     = ["Error-"]
xcelium_elab_log_error_regexes = ["*E "]
questa_elab_log_error_regexes  = ["\*\* Error:"]
riviera_elab_log_error_regexes = ["Error:"]

vivado_elab_log_warning_regexes  = ["WARNING:"]
metrics_elab_log_warning_regexes = ["=W:"]
vcs_elab_log_warning_regexes     = ["Warning-"]
xcelium_elab_log_warning_regexes = ["*W "]
questa_elab_log_warning_regexes  = ["\*\* Warning:"]
riviera_elab_log_warning_regexes = ["Warning:"]

sem = BoundedSemaphore(1)


def compile_fsoc_core(flist_path, core, sim_job):
    defines = sim_job.cmp_args
    timestamp_start = common.timestamp()
    var_name = "MIO_" + core.sname.replace("-", "_").upper() + "_SRC_PATH"
    os.environ[var_name] = core.dir
    sim_job.bwrap_flists[var_name] = "$PROJECT_ROOT_DIR/" + os.path.relpath(core.dir, cfg.project_dir)
    common.dbg(f"FuseSoC env var '{var_name}'='{core.dir}'")
    vendor = "@fsoc"
    log_file_path = compile_flist(vendor, core.sname, flist_path, [], sim_job, local=True)
    if not sim_job.dry_run:
        timestamp_end = common.timestamp()
        errors = scan_cmp_log_file_for_errors(log_file_path, sim_job)
        if len(errors):
            common.error("Errors during compilation of FuseSoC core '" + core.name + "':")
            for error in errors:
                common.error("  " + error)
            sim.kill_progress_bar()
            common.fatal("Stopping due to compilation.")
        log_cmp_history_fsoc(core, log_file_path, sim_job, timestamp_start, timestamp_end)
    return log_file_path


def compile_ip(ip, sim_job):
    ip_str = f"{ip.vendor}/{ip.name}"
    ip_dir = f"{ip.vendor}__{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    flist_path = get_ip_flist_path(ip, sim_job)
    deps_list  = get_dep_list(ip, sim_job)
    if sim_job.bwrap:
        if ip.is_global:
            if ip.is_encrypted:
                common.copy_directory(ip.path + "/" + ip.src_path + "." + sim_str, f"{cfg.temp_path}/{ip_dir}")
            else:
                common.copy_directory(ip.path + "/" + ip.src_path, f"{cfg.temp_path}/{ip_dir}")
            path = "${PROJECT_ROOT_DIR}/.mio/temp/" + ip_dir
        else:
            if ip.is_encrypted:
                path = "${PROJECT_ROOT_DIR}/" + os.path.relpath(ip.path + "/" + ip.src_path + "." + sim_str, cfg.project_dir)
            else:
                path = "${PROJECT_ROOT_DIR}/" + os.path.relpath(ip.path + "/" + ip.src_path, cfg.project_dir)
    elif ip.is_global and (sim_job.simulator == common.simulators_enum.METRICS):
        if ip.is_encrypted:
            common.copy_directory(ip.path + "/" + ip.src_path + "." + sim_str, f"{cfg.temp_path}/{ip_dir}")
            path = ".mio/temp/" + ip_dir
        else:
            common.copy_directory(ip.path + "/" + ip.src_path, f"{cfg.temp_path}/{ip_dir}")
            path = ".mio/temp/" + ip_dir
    else:
        if ip.is_encrypted:
            path = ip.path + "/" + ip.src_path + "." + sim_str
        else:
            path = ip.path + "/" + ip.src_path
    
    if (not sim_job.bwrap) and sim_job.simulator == common.simulators_enum.METRICS:
        set_uvm_version(sim_job)
        flist_path = os.path.relpath(flist_path, cfg.project_dir)
    
    if sim_job.bwrap:
        flist_path = os.path.relpath(flist_path, cfg.project_dir)
    
    flist_env_var_name = 'MIO_' + ip.name.upper() + '_SRC_PATH'
    os.environ[flist_env_var_name] = path
    sim_job.bwrap_flists[flist_env_var_name] = path
    timestamp_start = common.timestamp()
    log_file_path = compile_flist(ip.vendor, ip.name, flist_path, deps_list, sim_job, ip.is_local)
    if not sim_job.dry_run:
        timestamp_end = common.timestamp()
        errors = scan_cmp_log_file_for_errors(log_file_path, sim_job)
        if len(errors):
            common.error("Errors during compilation of IP '" + ip_str + "':")
            for error in errors:
                common.error("  " + error)
            sim.kill_progress_bar()
            common.fatal("Stopping due to compilation errors. Full log: " + log_file_path)
        log_cmp_history_ip(ip, log_file_path, sim_job, timestamp_start, timestamp_end)
        ip.is_compiled[sim_job.simulator] = True
    return log_file_path


def compile_vivado_project(ip, sim_job):
    ip_str = f"{ip.vendor}/{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    defines = sim_job.cmp_args
    dep_list = get_dep_list(ip, sim_job)
    timestamp_start = common.timestamp()
    log_file_paths = do_compile_vivado_project(ip, dep_list, sim_job)
    if not sim_job.dry_run:
        timestamp_end = common.timestamp()
        errors = scan_cmp_log_file_for_errors(log_file_paths[0], sim_job)
        errors.append(scan_cmp_log_file_for_errors(log_file_paths[1], sim_job))
        if len(errors):
            common.error("Errors during compilation of Vivado Project IP '" + ip_str + "':")
            for error in errors:
                common.error("  " + error)
            sim.kill_progress_bar()
            common.fatal("Stopping due to compilation errors. Logs: " + log_file_paths[0] + " & " + log_file_paths[0])
        log_cmp_history_vivado_project(ip, log_file_paths[0], log_file_paths[1], sim_job, timestamp_start, timestamp_end)
        ip.is_compiled[sim_job.simulator] = True
    return log_file_paths


def elaborate(ip, sim_job):
    ip_str = f"{ip.vendor}/{ip.name}"
    ip_dir_name = f"{ip.vendor}__{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    defines = sim_job.cmp_args
    if sim_job.is_regression:
        ip_dir_name = f"{ip.vendor}__{ip.name}__{sim_job.regression_name}"
        common.create_dir(cfg.sim_output_dir + "/" + sim_str + "/elab_out/regressions/" + ip_dir_name)
        common.create_dir(cfg.sim_output_dir + "/" + sim_str + "/elab_out/regressions/" + ip_dir_name + "/" + sim_job.regression_timestamp)
        elab_out =        cfg.sim_output_dir + "/" + sim_str + "/elab_out/regressions/" + ip_dir_name + "/" + sim_job.regression_timestamp
    else:
        ip_dir_name = f"{ip.vendor}__{ip.name}"
        elab_out = cfg.sim_output_dir + "/" + sim_str + "/elab_out/single_sim/" + ip_dir_name
    common.create_dir(elab_out)
    timestamp_start = common.timestamp()
    log_file_path = do_elaborate(ip, sim_job, elab_out)
    if not sim_job.dry_run:
        timestamp_end = common.timestamp()
        errors = scan_elab_log_file_for_errors(log_file_path, sim_job)
        if len(errors):
            common.error("Errors during elaboration of IP '" + ip_str + "':")
            for error in errors:
                common.error("  " + error)
            sim.kill_progress_bar()
            common.fatal("Stopping due to elaboration errors. Full log: " + log_file_path)
        log_elab_history(ip, log_file_path, sim_job, timestamp_start, timestamp_end)
        ip.is_elaborated[sim_job.simulator] = True
    return elab_out


def simulate(ip, sim_job):
    ip_str = f"{ip.vendor}/{ip.name}"
    ip_dir_name = f"{ip.vendor}__{ip.name}"
    sim_args = sim_job.sim_args
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    global sem
    start = common.timestamp()
    log_sim_start_history(ip, sim_job, start)
    if sim_job.is_regression:
        ip_dir_name = f"{ip.vendor}__{ip.name}__{sim_job.regression_name}"
        regr_sim_out_path = cfg.sim_output_dir + "/" + sim_str + "/elab_out/regressions/" + ip_dir_name + "/" + sim_job.regression_timestamp
        common.create_dir(regr_sim_out_path)
        sim_out = regr_sim_out_path + "/" + sim_str + "/elab_out/regressions/" + ip_dir_name + "/" + sim_job.regression_timestamp
        common.create_dir(sim_out)
    else:
        ip_dir_name = f"{ip.vendor}__{ip.name}"
        sim_out = cfg.sim_output_dir + "/" + sim_str + "/elab_out/single_sim/" + ip_dir_name
        common.create_dir(sim_out)
    do_simulate(ip, sim_job, sim_out)
    if not sim_job.dry_run:
        sem.acquire()
        log_sim_end_history(ip, sim_job, start, common.timestamp())
        sem.release()


def init_metrics_workspace():
    mdc_path = cfg.project_dir + "/.mdc"
    if not os.path.exists(mdc_path):
        common.info("Initializing Metrics workspace, this will take a few minutes ...")
        launch_eda_bin(cfg.metrics_home + "/mdc", ["initialize"], wd=cfg.project_dir)
        launch_eda_bin(cfg.metrics_home + "/mdc", ["status"], wd=cfg.project_dir, output=True)
        sync_ignore_filepath = f"{mdc_path}/sync_ignore"
        if os.path.exists(sync_ignore_filepath):
            with open(sync_ignore_filepath, "a") as f:
                src_dir_name = cfg.configuration.get("simulation", {}).get("root-path")
                f.writelines([".mio/sim", src_dir_name, "tools", "syn", "dft", "lint", "docs"])
        else:
            common.fatal(f"Failed to initialize Metrics Cloud Simulator workspace")


def compile_flist(vendor, name, flist_path, deps, sim_job, local, licensed=True):
    defines = sim_job.cmp_args
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    arg_list = []
    cmp_args_list = convert_compilation_args(sim_job)
    deps_list     = convert_deps_to_args(deps, sim_job)
    incdir_list   = get_incdir_list(deps, sim_job)
    
    #if licensed:
    #    license_ip = cache.get_ip("datum", "uvml_mio_lic", True)
    #    license_macros_file_path = license_ip.path + "/" + license_ip.src_path + "/uvml_mio_lic_macros.svh"
    license_macros_file_path = ""
    
    ip_dir_name = f"{vendor}__{name}"
    cmp_out_dir = cfg.sim_output_dir + "/" + sim_str + "/cmp_out/"
    cmp_out = cmp_out_dir + ip_dir_name
    sim_out = cfg.sim_output_dir + "/" + sim_str + "/cmp_wd/" + ip_dir_name
    common.create_dir(cmp_out)
    common.create_dir(sim_out)
    compilation_log_path = cfg.sim_dir + "/cmp/" + ip_dir_name + "." + sim_str + ".cmp.log"
    compilation_command_file = f"{ip_dir_name}.{sim_str}.cmp.cmd.txt"
    
    if sim_job.simulator == common.simulators_enum.VIVADO:
        arg_list += vivado_default_compilation_args
        arg_list += incdir_list
        arg_list.append(license_macros_file_path);
        arg_list.append("-f " + flist_path)
        arg_list += cmp_args_list
        arg_list.append("-L uvm")
        arg_list += deps_list
        arg_list.append(f"--work {name}={cmp_out}")
        arg_list.append("--log "  + compilation_log_path)
        write_cmd_to_disk(sim_job, "xvlog", arg_list, compilation_command_file)
        sim_job.bwrap_commands += launch_eda_bin(cfg.vivado_home + "/xvlog", arg_list, wd=sim_out, output=cfg.dbg, dry_run=sim_job.dry_run)
        
    elif sim_job.simulator == common.simulators_enum.VCS:
        arg_list += vcs_default_compilation_args
        arg_list.append(license_macros_file_path);
        arg_list.append("-f " + flist_path)
        arg_list += cmp_args_list
        arg_list += deps_list
        arg_list.append("-l "  + compilation_log_path)
        write_cmd_to_disk(sim_job, "vcs", arg_list, compilation_command_file)
        sim_job.bwrap_commands += launch_eda_bin(cfg.vcs_home + "/vcs", arg_list, wd=sim_out, output=cfg.dbg, dry_run=sim_job.dry_run)
        
    elif sim_job.simulator == common.simulators_enum.METRICS:
        arg_list += metrics_default_compilation_args
        arg_list += incdir_list
        if vendor == "@global":
            arg_list.append(f"-lib global__{name}")
        elif vendor == "@fsoc":
            arg_list.append(f"-lib fsoc__{name}")
        else:
            arg_list.append(f"-lib {vendor}__{name}")
        #common.copy_file(license_macros_file_path, f"{cfg.temp_path}/uvml_mio_lic_macros.svh")
        #arg_list.append(f".mio/temp/uvml_mio_lic_macros.svh") # HACK compute that path
        arg_list.append("-F " + flist_path)
        mtr_compilation_log_path = ip_dir_name + "." + sim_str + ".cmp.log"
        arg_list.append("-l " + mtr_compilation_log_path)
        arg_list.append(f"-timescale {cfg.sim_timescale}")
        
        arg_list_str = ""
        for arg in arg_list:
            arg_list_str = arg_list_str + f" {arg}"
        if cfg.dbg:
            arg_list = [f"dvlcom -v -a '{arg_list_str}'"]
        else:
            arg_list = [f"dvlcom -a '{arg_list_str}'"]
        #launch_eda_bin(cfg.metrics_home + "/mdc", ["initialize"], wd=cfg.project_dir, output=True) # TODO Add project.yml and store this in there
        write_cmd_to_disk(sim_job, "mdc", arg_list, compilation_command_file)
        sim_job.bwrap_commands += launch_eda_bin(cfg.metrics_home + "/mdc", arg_list, wd=cfg.project_dir, output=cfg.dbg, dry_run=sim_job.dry_run)
        launch_eda_bin(cfg.metrics_home + "/mdc", ["download", mtr_compilation_log_path], wd=cfg.project_dir, output=cfg.dbg, dry_run=sim_job.dry_run)
        if not sim_job.dry_run:
            common.move_file(f"{cfg.project_dir}/_downloaded_{mtr_compilation_log_path}", compilation_log_path)
        
    elif sim_job.simulator == common.simulators_enum.XCELIUM:
        arg_list += xcelium_default_compilation_args
        arg_list.append(license_macros_file_path);
        arg_list.append("-f " + flist_path)
        # TODO Add compilation output argument for nc
        write_cmd_to_disk(sim_job, "xrun", arg_list, compilation_command_file)
        sim_job.bwrap_commands += launch_eda_bin(cfg.nc_home + "/xrun", arg_list, wd=sim_out, output=cfg.dbg, dry_run=sim_job.dry_run)
        
    elif sim_job.simulator == common.simulators_enum.QUESTA:
        os.environ['MIO_UVM_HOME'] = f"$MIO_QUESTA_HOME/../verilog_src/uvm-{cfg.uvm_version}"
        arg_list += questa_default_compilation_args
        arg_list.append(license_macros_file_path);
        arg_list.append("-f " + flist_path)
        arg_list += cmp_args_list
        arg_list += deps_list
        arg_list.append(f"-Ldir {cmp_out_dir}")
        arg_list.append("-l "  + compilation_log_path)
        arg_list.append(f"-work {name}")
        write_cmd_to_disk(sim_job, "vlog", arg_list, compilation_command_file)
        sim_job.bwrap_commands += launch_eda_bin(cfg.questa_home + "/vlog", arg_list, wd=sim_out, output=cfg.dbg)
        
    elif sim_job.simulator == common.simulators_enum.RIVIERA:
        arg_list += riviera_default_compilation_args
        arg_list.append(license_macros_file_path);
        arg_list.append("-f " + flist_path)
        # TODO Add compilation output argument for riviera
        write_cmd_to_disk(sim_job, "vlog", arg_list, compilation_command_file)
        sim_job.bwrap_commands += launch_eda_bin(cfg.riviera_home + "/vlog", arg_list, wd=sim_out, output=cfg.dbg, dry_run=sim_job.dry_run)
    
    return compilation_log_path


def do_compile_vivado_project(ip, deps, sim_job, local):
    ip_str = f"{ip.vendor}/{ip.name}"
    ip_dir_name = f"{ip.vendor}__{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    cmp_out = cfg.sim_output_dir + "/" + sim_str + "/cmp_out/" + ip_dir_name
    sim_out = cfg.sim_output_dir + "/" + sim_str + "/cmp_wd/" + ip_dir_name
    
    if sim_job.simulator == common.simulators_enum.VIVADO:
        vlog_compilation_log_path = cfg.sim_dir + "/cmp/" + ip_dir_name + ".viv.vlog.cmp.log"
        vhdl_compilation_log_path = cfg.sim_dir + "/cmp/" + ip_dir_name + ".viv.vhdl.cmp.log"
        vlog_arg_list = ["-prj "  + ip.vproj_vlog]
        vlog_arg_list.append("--work " + ip.name + "=" + cmp_out)
        vlog_arg_list += vivado_project_default_vlog_compilation_args
        vlog_arg_list += convert_defines(sim_job)
        vlog_arg_list += convert_compilation_args(sim_job)
        vlog_arg_list += convert_deps_to_args(deps, sim_job)
        vlog_arg_list.append("--log " + vlog_compilation_log_path)
        vhdl_arg_list = ["-prj " + ip.vproj_vhdl]
        vhdl_arg_list.append("--work " + ip.name + "=" + cmp_out)
        vhdl_arg_list += vivado_project_default_vhdl_compilation_args
        vlog_arg_list += convert_defines(sim_job)
        vlog_arg_list += convert_compilation_args(sim_job)
        vlog_arg_list += convert_deps_to_args(deps, sim_job)
        vhdl_arg_list.append("--log " + vhdl_compilation_log_path)
        compilation_command_file = f"{ip_dir_name}.{sim_str}.cmp.xvlog.cmd.txt"
        write_cmd_to_disk(sim_job, "xvlog", arg_list, compilation_command_file)
        sim_job.bwrap_commands += launch_eda_bin(cfg.vivado_home + "/xvlog", vlog_arg_list, wd=sim_out, dry_run=sim_job.dry_run)
        compilation_command_file = f"{ip_dir_name}.{sim_str}.cmp.xvhdl.cmd.txt"
        write_cmd_to_disk(sim_job, "xvhdl", arg_list, compilation_command_file)
        sim_job.bwrap_commands += launch_eda_bin(cfg.vivado_home + "/xvhdl", vhdl_arg_list, wd=sim_out, dry_run=sim_job.dry_run)
    else:
        common.fatal("Vivado Project IP are not yet compatible with simulator '" + sim_str + "'.")
    
    return [vlog_compilation_log_path, vhdl_compilation_log_path]


def do_elaborate(ip, sim_job, wd):
    ip_str = f"{ip.vendor}/{ip.name}"
    ip_dir_name = f"{ip.vendor}__{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    arg_list = []
    def_list  = convert_defines(sim_job)
    elab_list = convert_elaboration_args(sim_job)
    deps_ip_list = get_dep_list(ip, sim_job)
    deps_list = convert_deps_to_args(deps_ip_list, sim_job)
    so_libs = get_all_so_libs(ip, sim_job)
    elaboration_log_path = cfg.sim_dir + "/elab/" + ip_dir_name + "." + sim_str + ".elab.log"
    cmp_out_dir = cfg.sim_output_dir + "/" + sim_str + "/cmp_out/"
    ip_cmp_path = cmp_out_dir + ip_dir_name
    elaboration_command_file = f"{ip_dir_name}.{sim_str}.elab.cmd.txt"
    
    if sim_job.simulator == common.simulators_enum.VIVADO:
        if ip.has_dut:
            arg_list += dut_elab_to_arg_list(ip, sim_job)
        arg_list += def_list
        arg_list += elab_list
        arg_list += deps_list
        arg_list += vivado_default_elaboration_args
        arg_list.append("--log "  + elaboration_log_path)
        arg_list.append("-s "     + ip.name)
        arg_list.append("-L "     + ip.name + "=" + ip_cmp_path)
        for construct in ip.hdl_src_top_constructs:
            if "." in construct:
                arg_list.append(construct)
            else:
                arg_list.append(ip.name + "." + construct)
        
        common.create_dir(wd)
        arg_list.append(f"-sv_root {wd}")
        
        for so_lib in so_libs:
            common.copy_file(so_libs[so_lib], f"{wd}/{so_lib}")
            arg_list.append(f"-sv_lib {so_lib}")
        
        write_cmd_to_disk(sim_job, "xelab", arg_list, elaboration_command_file)
        sim_job.bwrap_commands += launch_eda_bin(cfg.vivado_home + "/xelab", arg_list, wd, output=cfg.dbg, dry_run=sim_job.dry_run)
        
    elif sim_job.simulator == common.simulators_enum.VCS:
        arg_list += vcs_default_elaboration_args
        if ip.has_dut:
            arg_list += dut_elab_to_arg_list(ip, sim_job)
        # TODO Add elaboration output argument for vcs
        write_cmd_to_disk(sim_job, "vcs", arg_list, elaboration_command_file)
        sim_job.bwrap_commands += launch_eda_bin(cfg.vcs_home + "/vcs", arg_list, wd, output=cfg.dbg, dry_run=sim_job.dry_run)
        
    elif sim_job.simulator == common.simulators_enum.METRICS:
        arg_list += metrics_default_elaboration_args
        mtr_elaboration_log_path = ip_dir_name + "." + sim_str + ".elab.log"
        arg_list.append(f"-l {mtr_elaboration_log_path}")
        arg_list.append(f"-timescale {cfg.sim_timescale}")
        arg_list += def_list
        arg_list += elab_list
        arg_list += deps_list
        if ip.has_dut:
            arg_list += dut_elab_to_arg_list(ip, sim_job)
        #arg_list.append(f"+incdir+%UVM_HOME%/src")
        #arg_list.append(f"%UVM_HOME%/src/uvm_pkg.sv")
        
        if ip.vendor == "@global":
            arg_list.append(f"-genimage global__{ip.name}")
            arg_list.append(f"-L global__{ip.name}")
        elif ip.vendor == "@fsoc":
            arg_list.append(f"-genimage fsoc__{ip.name}")
            arg_list.append(f"-L fsoc__{ip.name}")
        else:
            arg_list.append(f"-genimage {ip.vendor}__{ip.name}")
            arg_list.append(f"-L {ip.vendor}__{ip.name}")
        
        for construct in ip.hdl_src_top_constructs:
            if "." in construct:
                lib,name = construct.split(".")
                arg_list.append(f"-top {name}")
            else:
                arg_list.append(f"-top {construct}")
        
        arg_list_str = ""
        for arg in arg_list:
            arg_list_str = arg_list_str + f" {arg}"
        if cfg.dbg:
            arg_list = [f"dsim -v -a '{arg_list_str}'"]
        else:
            arg_list = [f"dsim -a '{arg_list_str}'"]
        write_cmd_to_disk(sim_job, "mdc", arg_list, elaboration_command_file)
        sim_job.bwrap_commands += launch_eda_bin(cfg.metrics_home + "/mdc", arg_list, wd=cfg.project_dir, output=cfg.dbg, dry_run=sim_job.dry_run)
        sim_job.bwrap_commands += launch_eda_bin(cfg.metrics_home + "/mdc", ["download", mtr_elaboration_log_path], wd=cfg.project_dir, output=cfg.dbg, dry_run=sim_job.dry_run)
        if not sim_job.dry_run:
            common.move_file(f"{cfg.project_dir}/_downloaded_{mtr_elaboration_log_path}", elaboration_log_path)
        
    elif sim_job.simulator == common.simulators_enum.XCELIUM:
        arg_list += xcelium_default_elaboration_args
        if ip.has_dut:
            arg_list += dut_elab_to_arg_list(ip, sim_job)
        # TODO Add elaboration output argument for nc
        write_cmd_to_disk(sim_job, "xrun", arg_list, elaboration_command_file)
        sim_job.bwrap_commands += launch_eda_bin(cfg.nc_home + "/xrun", arg_list, wd, output=cfg.dbg, dry_run=sim_job.dry_run)
        
    elif sim_job.simulator == common.simulators_enum.QUESTA:
        for construct in ip.hdl_src_top_constructs:
            if "." in construct:
                arg_list.append(construct)
            else:
                arg_list.append(ip.name + "." + construct)
        
        arg_list += questa_default_elaboration_args
        if ip.has_dut:
            arg_list += dut_elab_to_arg_list(ip, sim_job)
        arg_list += def_list
        arg_list += elab_list
        arg_list += deps_list
        arg_list.append(f"-work {ip_dir_name}")
        arg_list.append(f"-o {ip.name}")
        arg_list.append(f"-l {elaboration_log_path}")
        arg_list.append(f"-Ldir {cmp_out_dir}")
        write_cmd_to_disk(sim_job, "vopt", arg_list, elaboration_command_file)
        sim_job.bwrap_commands += launch_eda_bin(cfg.questa_home + "/vopt", arg_list, wd, output=cfg.dbg, dry_run=sim_job.dry_run)
        
    elif sim_job.simulator == common.simulators_enum.RIVIERA:
        arg_list += riviera_default_elaboration_args
        if ip.has_dut:
            arg_list += dut_elab_to_arg_list(ip, sim_job)
        # TODO Add elaboration output argument for riviera
        write_cmd_to_disk(sim_job, "vlog", arg_list, elaboration_command_file)
        sim_job.bwrap_commands += launch_eda_bin(cfg.riviera_home + "/vlog", arg_list, wd, output=cfg.dbg, dry_run=sim_job.dry_run)
    
    return elaboration_log_path


def do_simulate(ip, sim_job, wd):
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    ip_str = f"{ip.vendor}/{ip.name}"
    ip_dir_name = f"{ip.vendor}__{ip.name}"
    global sem
    sem.acquire()
    plus_args = sim_job.sim_args
    if (len(plus_args) > 0):
        args_present = True
    else:
        args_present = False
    arg_list = []
    test_template = Template(ip.hdl_src_test_name_template)
    test_result_dir_template = Template(cfg.test_results_path_template)
    test = sim_job.test
    test_name = test_template.render(name=test)
    simulation_command_file = f"{ip_dir_name}.{sim_str}.sim.cmd.txt"
    test_result_dir = test_result_dir_template.render(ip_vendor=ip.vendor, ip_name=ip.name, test_name=test, seed=sim_job.seed, args=plus_args_list_to_str_list(plus_args), args_present=args_present)
    plus_args["UVM_TESTNAME"] = test_name
    
    plus_args["__MIO_TOKEN"] = user.login()
    
    if sim_job.is_regression:
        common.create_dir   (cfg.regr_results_dir + "/" + ip.name + "_" + sim_job.regression_name)
        common.create_dir   (cfg.regr_results_dir + "/" + ip.name + "_" + sim_job.regression_name + "/" + sim_job.regression_timestamp)
        tests_results_path = cfg.regr_results_dir + "/" + ip.name + "_" + sim_job.regression_name + "/" + sim_job.regression_timestamp + "/" + test_result_dir
    else:
        tests_results_path = cfg.sim_results_dir + "/" + test_result_dir
    
    output = not sim_job.is_regression
    
    simulation_log_path = tests_results_path + "/sim.log"
    cov_path = tests_results_path + "/cov"
    common.create_dir(tests_results_path)
    common.create_dir(tests_results_path + "/trn_log")
    sim_job.results_path = tests_results_path
    sim_job.results_dir_name = test_result_dir
    if sim_job.waves:
        if sim_job.simulator == common.simulators_enum.VIVADO:
            wave_capture_script_path = tests_results_path + "/waves.viv.tcl"
            waves_path = tests_results_path + "/waves.wdb"
            arg_list.append("--wdb "      + waves_path              )
            arg_list.append("--tclbatch " + wave_capture_script_path)
            if not os.path.exists(wave_capture_script_path):
                try:
                    common.create_file(wave_capture_script_path)
                    f = open(wave_capture_script_path, "w")
                    f.write("log_wave -recursive * \n")
                    f.write("run -all \n")
                    f.write("quit \n")
                    f.close()
                except Exception as e:
                    common.fatal("Could not create wave capture script at " + wave_capture_script_path + ": " + str(e))
        elif sim_job.simulator == common.simulators_enum.VCS:
            # TODO Implement waves file for vcs
            pass
        elif sim_job.simulator == common.simulators_enum.METRICS:
            arg_list.append(f"-waves {test_result_dir}.vcd")
        elif sim_job.simulator == common.simulators_enum.XCELIUM:
            # TODO Implement waves file for xcelium
            pass
        elif sim_job.simulator == common.simulators_enum.QUESTA:
            # TODO Implement waves file for questa
            pass
        elif sim_job.simulator == common.simulators_enum.RIVIERA:
            # TODO Implement waves file for riviera
            pass
    
    if sim_job.cov:
        common.create_dir(cov_path)
        if sim_job.simulator == common.simulators_enum.VIVADO:
            arg_list.append("-cov_db_name " + test_name)
            arg_list.append("-cov_db_dir "  + cov_path )
        elif sim_job.simulator == common.simulators_enum.VCS:
            # TODO Implement coverage for vcs
            pass
        elif sim_job.simulator == common.simulators_enum.METRICS:
            arg_list.append(f"-code-cov a")
            arg_list.append(f"-cov-db {test_result_dir}")
        elif sim_job.simulator == common.simulators_enum.XCELIUM:
            # TODO Implement coverage for xcelium
            pass
        elif sim_job.simulator == common.simulators_enum.QUESTA:
            # TODO Implement coverage for questa
            pass
        elif sim_job.simulator == common.simulators_enum.RIVIERA:
            # TODO Implement coverage for riviera
            pass
    
    plus_args["UVM_NO_RELNOTES"                ] = ""
    plus_args["SIM_DIR_RESULTS"                ] = cfg.sim_results_dir
    plus_args["UVM_VERBOSITY"                  ] = "UVM_" + sim_job.verbosity.upper()
    plus_args["UVM_MAX_QUIT_COUNT"             ] = str(sim_job.max_errors)
    
    if sim_job.simulator == common.simulators_enum.METRICS:
        plus_args["UVMX_FILE_BASE_DIR_SIM"         ] = "./.mio/temp"
        plus_args["UVMX_FILE_BASE_DIR_TEST_RESULTS"] = "./.mio/temp"
        common.create_dir(cfg.temp_path + "/trn_log")
        plus_args["UVMX_FILE_BASE_DIR_TB"          ] = os.path.relpath(ip.path + "/" + ip.src_path, cfg.project_dir)
        plus_args["UVMX_FILE_BASE_DIR_TESTS"       ] = os.path.relpath(ip.path + "/" + ip.src_path + "/" + ip.hdl_src_tests_path, cfg.project_dir)
    else:
        plus_args["UVMX_FILE_BASE_DIR_SIM"         ] = cfg.sim_dir
        plus_args["UVMX_FILE_BASE_DIR_TEST_RESULTS"] = tests_results_path
        plus_args["UVMX_FILE_BASE_DIR_TB"          ] = ip.path + "/" + ip.src_path
        plus_args["UVMX_FILE_BASE_DIR_TESTS"       ] = ip.path + "/" + ip.src_path + "/" + ip.hdl_src_tests_path
    
    plus_args_list = convert_plus_args(sim_job)
    if sim_job.simulator == common.simulators_enum.VIVADO:
        arg_list += plus_args_list
        arg_list += vivado_default_simulation_args
        arg_list.append("--log " + simulation_log_path)
        if sim_job.gui:
            arg_list.append("--gui")
        else:
            if not sim_job.waves:
                arg_list.append("--runall")
                arg_list.append("--onerror quit")
        if not sim_job.cov:
            arg_list.append("-ignore_coverage")
        arg_list.append(ip.name)
        arg_list.append("-sv_seed " + str(sim_job.seed))
        write_cmd_to_disk(sim_job, "xsim", arg_list, simulation_command_file)
        sem.release()
        sim_job.bwrap_commands += launch_eda_bin(cfg.vivado_home + "/xsim", arg_list, wd, output=output, dry_run=sim_job.dry_run)
        
    elif sim_job.simulator == common.simulators_enum.VCS:
        arg_list += vcs_default_simulation_args
        # TODO Add simulation output argument for vcs
        write_cmd_to_disk(sim_job, "simv", arg_list, simulation_command_file)
        sem.release()
        sim_job.bwrap_commands += launch_eda_bin(cfg.vcs_home + "/simv", arg_list, wd, output=output, dry_run=sim_job.dry_run)
        
    elif sim_job.simulator == common.simulators_enum.METRICS:
        arg_list += metrics_default_simulation_args
        arg_list += plus_args_list
        mtr_simulation_log_path = f"{test_result_dir}.log"
        arg_list.append("-l " + mtr_simulation_log_path)
        arg_list.append("-sv_seed " + str(sim_job.seed))
        arg_list.append(f"-image {ip.vendor}__{ip.name}")
        arg_list.append(f"-sv_lib %UVM_HOME%/src/dpi/libuvm_dpi.so")
        arg_list.append(f"-timescale {cfg.sim_timescale}")
        #arg_list.append(f"-work {ip.vendor}__{ip.name}")
        
        arg_list_str = ""
        for arg in arg_list:
            arg_list_str = arg_list_str + f" {arg}"
        if cfg.dbg:
            arg_list = [f"dsim -v -a '{arg_list_str}'"]
        else:
            arg_list = [f"dsim -a '{arg_list_str}'"]
        write_cmd_to_disk(sim_job, "mdc", arg_list, simulation_command_file)
        sem.release()
        sim_job.bwrap_commands += launch_eda_bin(cfg.metrics_home + "/mdc", arg_list, wd=cfg.project_dir, output=output, dry_run=sim_job.dry_run)
        common.remove_file(f"{cfg.project_dir}/_downloaded_{mtr_simulation_log_path}")
        sim_job.bwrap_commands += launch_eda_bin(cfg.metrics_home + "/mdc", ["download", mtr_simulation_log_path], wd=cfg.project_dir, dry_run=sim_job.dry_run)
        if not sim_job.dry_run:
            common.move_file(f"{cfg.project_dir}/_downloaded_{mtr_simulation_log_path}", simulation_log_path)
            if sim_job.waves:
                common.remove_file(f"{cfg.project_dir}/_downloaded_{test_result_dir}.vcd")
                sim_job.bwrap_commands += launch_eda_bin(cfg.metrics_home + "/mdc", ["download", f"{test_result_dir}.vcd"], wd=cfg.project_dir, dry_run=sim_job.dry_run)
                common.move_file(f"{cfg.project_dir}/_downloaded_{test_result_dir}.vcd", f"{tests_results_path}/waves.vcd")
        
    elif sim_job.simulator == common.simulators_enum.XCELIUM:
        arg_list += xcelium_default_simulation_args
        # TODO Add simulation output argument for nc
        write_cmd_to_disk(sim_job, "xrun", arg_list, simulation_command_file)
        sem.release()
        sim_job.bwrap_commands += launch_eda_bin(cfg.nc_home + "/xrun", arg_list, wd, output=output, dry_run=sim_job.dry_run)
        
    elif sim_job.simulator == common.simulators_enum.QUESTA:
        arg_list += questa_default_simulation_args
        arg_list += plus_args_list
        arg_list.append("-l " + simulation_log_path)
        arg_list.append("-sv_seed " + str(sim_job.seed))
        arg_list.append(f" {ip.name}")
        write_cmd_to_disk(sim_job, "vsim", arg_list, simulation_command_file)
        sem.release()
        sim_job.bwrap_commands += launch_eda_bin(cfg.questa_home + "/vsim", arg_list, wd, output=output, dry_run=sim_job.dry_run)
        
    elif sim_job.simulator == common.simulators_enum.RIVIERA:
        arg_list += riviera_default_simulation_args
        # TODO Add simulation output argument for riviera
        write_cmd_to_disk(sim_job, "vsim", arg_list, simulation_command_file)
        sem.release()
        sim_job.bwrap_commands += launch_eda_bin(cfg.riviera_home + "/vsim", arg_list, wd, output=output, dry_run=sim_job.dry_run)
    
    sim_job.sim_log_file_path = simulation_log_path


def write_cmd_to_disk(sim_job, executable_name, arg_list, command_filename):
    arg_list_str = ""
    for arg in arg_list:
        arg_list_str = f"{arg_list_str} {arg}"
    if sim_job.simulator == common.simulators_enum.VIVADO:
        command_str = f"$MIO_VIVADO_HOME/{executable_name}{arg_list_str}"
    elif sim_job.simulator == common.simulators_enum.METRICS:
        command_str = f"$MIO_METRICS_HOME/{executable_name}{arg_list_str}"
    elif sim_job.simulator == common.simulators_enum.VCS:
        command_str = f"$MIO_VCS_HOME/{executable_name}{arg_list_str}"
    elif sim_job.simulator == common.simulators_enum.QUESTA:
        command_str = f"$MIO_QUESTA_HOME/{executable_name}{arg_list_str}"
    elif sim_job.simulator == common.simulators_enum.XCELIUM:
        command_str = f"$MIO_XCELIUM_HOME/{executable_name}{arg_list_str}"
    elif sim_job.simulator == common.simulators_enum.RIVIERA:
        command_str = f"$MIO_RIVIERA_HOME/{executable_name}{arg_list_str}"
    
    try:
        path = cfg.temp_path + "/" + command_filename
        cmd_file = open(path, 'w')
        cmd_file.write(command_str)
        cmd_file.close()
    except Exception as e:
        common.fatal(f"Could not write command to disk: {e}")


def set_uvm_version(sim_job):
    found_uvm_ver = False
    if sim_job.simulator == common.simulators_enum.METRICS:
        config_path = cfg.project_dir + "/.mdc/config.yml"
        try:
            with open(config_path, 'r') as cfg_file:
                cfg_yaml = yaml.load(cfg_file, Loader=SafeLoader)
                if 'toolchain' not in cfg_yaml:
                    cfg_yaml['toolchain'] = {}
                if 'packages' not in cfg_yaml['toolchain']:
                    cfg_yaml['toolchain']['packages'] = []
                for pkg in cfg_yaml['toolchain']['packages']:
                    if pkg['name'] == 'uvm':
                        found_uvm_ver = True
                        pkg['version'] = cfg.uvm_version
                        break
                if not found_uvm_ver:
                    cfg_yaml['toolchain']['packages'].append({"name":"uvm", "version":cfg.uvm_version})
            with open(config_path, 'w') as cfg_file_write:
                yaml.dump(cfg_yaml, cfg_file_write)
        except Exception as e:
            common.fatal(f"Failed to set UVM version for Metrics simulator: {e}")


def dut_elab_to_arg_list(ip, sim_job):
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    args = []
    if (ip.dut_ip_type == "fsoc"):
        file_path_partial_name = re.sub(r':', '_', ip.dut_fsoc_full_name)
        eda_file_dir = cfg.fsoc_dir + "/" + ip.dut_fsoc_name + "/sim-xsim"
        eda_file_path = eda_file_dir + "/" + file_path_partial_name + "_0.eda.yml"
        try:
            if os.path.exists(eda_file_path):
                with open(eda_file_path, 'r') as edafile:
                    eda_yaml = yaml.load(edafile, Loader=SafeLoader)
                    if eda_yaml:
                        elab_options = eda_yaml['tool_options']['xsim']['xelab_options']
                        if sim_job.simulator == common.simulators_enum.METRICS:
                            args.append("-L " + "fsoc__" + ip.dut_fsoc_name)
                        else:
                            args.append("-L " + ip.dut_fsoc_name + "=" + cfg.sim_output_dir + "/" + sim_str + "/cmp_out/@fsoc__" + ip.dut_fsoc_name)
                    else:
                        common.fatal("ERROR: Unable to parse FuseSoC output " + eda_file_path)
            else:
                common.fatal("ERROR: Unable to parse FuseSoC output " + eda_file_path)
        except Exception as e:
            common.fatal("ERROR: Unable to find FuseSoC core output for '" + ip.dut_fsoc_name + "': " + str(e))
    else:
        dut_ip = ip.dut.target_ip_model
        dut_ip_dir_name = f"{dut_ip.vendor}__{dut_ip.name}"
        if dut_ip.sub_type == "vivado":
            if sim_job.simulator != common.simulators_enum.VIVADO:
                common.fatal("Vivado Projects are currently only compatible with the Vivado simulator.")
            for construct in dut_ip.hdl_src_top_constructs:
                args.append(construct)
            for lib in dut_ip.vproj_libs:
                args.append("-L " + lib)
        else:
            # Performed by get_dep_list()
            pass
            #if sim_job.simulator == common.simulators_enum.METRICS:
            #    if dut_ip.vendor == "@global":
            #        lib_str = f"-L global__{dut_ip.name}"
            #    else:
            #        lib_str = f"-L {dut_ip.vendor}__{dut_ip.name}"
            #elif sim_job.simulator == common.simulators_enum.QUESTA:
            #    lib_str = f"-L {dut_ip.vendor}__{dut_ip.name}"
            #else:
            #    lib_str = "-L " + dut_ip.name + "=" + cfg.sim_output_dir + "/" + sim_str + "/cmp_out/" + dut_ip_dir_name
            #args.append(lib_str)
    common.dbg("DUT args list: " + str(args))
    return args


def get_ip_flist_path(ip, sim_job):
    ip_str = f"{ip.vendor}/{ip.name}"
    ip_dir = f"{ip.vendor}__{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    found_flist = False
    flist_path = ""
    
    if ip.hdl_src_flists[sim_job.simulator] != "":
        found_flist = True
        if ip.is_global and (sim_job.simulator == common.simulators_enum.METRICS):
            flist_path = f".mio/temp/{ip_dir}/{ip.hdl_src_flists[sim_job.simulator]}"
        else:
            if ip.is_encrypted:
                flist_path = f"{ip.path}/{ip.src_path}.{sim_str}/{ip.hdl_src_flists[sim_job.simulator]}"
            else:
                flist_path = f"{ip.path}/{ip.src_path}/{ip.hdl_src_flists[sim_job.simulator]}"
    if not found_flist:
        flist_path = gen_flist(ip, sim_job)
    return flist_path


def gen_master_flist(ip, sim_job, deps):
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    flists = gen_flists(ip, deps, sim_job)
    flist_path = cfg.temp_path + "/" + ip.vendor + "__" + ip.name + ".top." + sim_str + ".flist"
    
    if sim_job.simulator == common.simulators_enum.METRICS:
        final_flists = []
        for flist in flists:
            final_flist = os.path.relpath(flist, cfg.temp_path)
            final_flists.append(final_flist)
        flists = final_flists
    
    gen_flist_file(sim_job.simulator, ip.vendor, ip.name, flist_path, [], flists, [], [])
    
    if sim_job.simulator == common.simulators_enum.METRICS:
        flist_path = os.path.relpath(flist_path, cfg.project_dir)
        flist_path = flist_path.replace(cfg.project_dir, "")
    
    return flist_path


def gen_flists(ip, deps, sim_job):
    flists = []
    for dep in deps:
        flists.append(get_ip_flist_path(dep, sim_job))
    flists.append(get_ip_flist_path(ip, sim_job))
    return flists


def gen_flist(ip, sim_job):
    ip_str = f"{ip.vendor}/{ip.name}"
    ip_dir = f"{ip.vendor}__{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    flist_path = ""
    defines = convert_defines(sim_job)
    if len(ip.hdl_src_top_files) == 0:
        common.fatal(f"No 'top-files' defined under section 'hdl-src' in descriptor for IP '{ip_str}'")
    if len(ip.hdl_src_directories) == 0:
        common.fatal(f"No 'directories' entry under section 'hdl-src' in descriptor for IP '{ip_str}'")
    
    rel_ip_path = os.path.relpath(ip.path, cfg.temp_path)
    simulator = sim_str
    flist_template = cfg.templateEnv.get_template(f"{simulator}.flist.j2")
    directories = []
    for dir in ip.hdl_src_directories:
        if dir == ".":
            if sim_job.simulator == common.simulators_enum.METRICS:
                if ip.is_global:
                    directories.append(f"{ip_dir}")
                else:
                    if ip.is_encrypted:
                        directories.append(f"{rel_ip_path}/{ip.src_path}.{sim_str}")
                    else:
                        directories.append(f"{rel_ip_path}/{ip.src_path}")
            else:
                directories.append("${MIO_" + ip.name.upper() + "_SRC_PATH}")
        else:
            if sim_job.simulator == common.simulators_enum.METRICS:
                if ip.is_global:
                    directories.append(f"{ip_dir}/{dir}")
                else:
                    if ip.is_encrypted:
                        directories.append(f"{rel_ip_path}/{ip.src_path}.{sim_str}/{dir}")
                    else:
                        directories.append(f"{rel_ip_path}/{ip.src_path}/{dir}")
            else:
                directories.append("${MIO_" + ip.name.upper() + "_SRC_PATH}/" + dir)
    
    top_files = []
    for file in ip.hdl_src_top_files:
        if sim_job.simulator == common.simulators_enum.METRICS:
            if ip.is_global:
                top_files.append(f"{ip_dir}/{file}")
            else:
                if ip.is_encrypted:
                    top_files.append(f"{rel_ip_path}/{ip.src_path}.{sim_str}/{file}")
                else:
                    top_files.append(f"{rel_ip_path}/{ip.src_path}/{file}")
        else:
            top_files.append("${MIO_" + ip.name.upper() + "_SRC_PATH}/" + file)
    
    flist_path = cfg.temp_path + "/" + ip.vendor + "__" + ip.name + "." + simulator + ".flist"
    if ip.type == "dv":
        include_uvm = True
    else:
        include_uvm = False
    gen_flist_file(sim_job.simulator, ip.vendor, ip.name, flist_path, defines, [], directories, top_files, include_uvm)
    common.dbg(f"Using filelist '{flist_path}' for IP '{ip_str}'")
    
    return flist_path


def gen_flist_file(simulator, ip_vendor, ip_name, path, defines, filelists, directories, files, include_uvm):
    ip_str = f"{ip_vendor}/{ip_name}"
    sim_str = common.get_simulator_short_name(simulator)
    
    if include_uvm:
        if simulator == common.simulators_enum.METRICS:
            directories.insert(0, "$UVM_HOME/src")
            files      .insert(0, "$UVM_HOME/src/uvm_pkg.sv")
        if simulator == common.simulators_enum.QUESTA:
            directories.insert(0, "$(MIO_UVM_HOME)/src")
            files      .insert(0, "$(MIO_UVM_HOME)/src/uvm_pkg.sv")
    
    try:
        flist_template = cfg.templateEnv.get_template(f"{sim_str}.flist.j2")
        common.dbg(f"Generating filelist for IP '{ip_str}' and simulator '{sim_str}' with files='{files}' and dirs='{directories}'")
        outputText = flist_template.render(defines=defines, filelists=filelists, files=files, dirs=directories)
        with open(path,'w') as flist_file:
            flist_file.write(outputText)
            flist_file.close()
    except Exception as e:
        common.fatal(f"Failed to create filelist for IP '{ip_str}': {e}")


def convert_defines(sim_job):
    defines = sim_job.cmp_args
    args = []
    for define in defines:
        if sim_job.simulator == common.simulators_enum.VIVADO:
            if defines[define] != "":
                args.append("--define " + define + "=" + defines[define])
            else:
                args.append("--define " + define)
        else:
            if defines[define] != "":
                args.append("+define+" + define + "=" + defines[define])
            else:
                args.append("+define+" + define)
    return args


def convert_plus_args(sim_job):
    plus_args = sim_job.sim_args
    args = []
    for arg in plus_args:
        if sim_job.simulator == common.simulators_enum.VIVADO:
            if plus_args[arg] != "":
                args.append("-testplusarg \"" + arg + "=" + plus_args[arg] + "\"")
            else:
                args.append("-testplusarg \"" + arg + "\"")
        else:
            if plus_args[arg] != "":
                args.append("+" + arg + "=" + plus_args[arg])
            else:
                args.append("+" + arg)
    return args


def convert_deps_to_args(deps, sim_job):
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    args = []
    for dep in deps:
        if dep.name != "uvm":
            if sim_job.simulator == common.simulators_enum.METRICS:
                if dep.vendor == "@global":
                    dep_dir_name = f"global__{dep.name}"
                elif dep.vendor == "@fsoc":
                    dep_dir_name = f"fsoc__{dep.name}"
                else:
                    dep_dir_name = f"{dep.vendor}__{dep.name}"
                lib_str = f"-L {dep_dir_name}"
            elif sim_job.simulator == common.simulators_enum.QUESTA:
                dep_dir_name = f"{dep.vendor}__{dep.name}"
                lib_str = f"-L {dep_dir_name}"
            else:
                dep_dir_name = f"{dep.vendor}__{dep.name}"
                lib_str = "-L " + dep.name + "=" + cfg.sim_output_dir + "/" + sim_str + "/cmp_out/" + dep_dir_name
            args.append(lib_str)
    return args


def get_dep_list(ip, sim_job):
    final_dep_list = []
    if ip.has_dut:
        if ip.dut_ip_type == "":
            final_dep_list.append(ip.dut.target_ip_model)
    dep_list = ip.get_ordered_deps()
    for dep in dep_list:
        if dep.name == "uvm": # UVM is considered part of the simulator
            continue
        final_dep_list.append(dep)
    return final_dep_list


def get_incdir_list(deps, sim_job):
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    incdir_list = []
    for dep in deps:
        for dir in dep.hdl_src_directories:
            if dep.is_global and (sim_job.bwrap or sim_job.simulator == common.simulators_enum.METRICS):
                src_path = f"{cfg.temp_path}/{dep.vendor}__{dep.name}/{dir}"
            else:
                if dep.is_encrypted:
                    src_path = f"{dep.path}/{dep.src_path}.{sim_str}/{dir}"
                else:
                    src_path = f"{dep.path}/{dep.src_path}/{dir}"
            if sim_job.simulator == common.simulators_enum.VIVADO:
                incdir_list.append("-i " + src_path)
            else:
                if (sim_job.bwrap or sim_job.simulator == common.simulators_enum.METRICS):
                    src_path = os.path.relpath(src_path, cfg.project_dir)
                incdir_list.append("+incdir+" + src_path)
    return incdir_list


def get_all_so_libs(ip, sim_job):
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    so_libs = {}
    for dep in ip.dependencies:
        dep_so_libs = get_all_so_libs(dep.target_ip_model, sim_job)
        common.merge_dict(so_libs, dep_so_libs)
    for so_lib in ip.hdl_src_so_libs:
        so_lib_flat_name = f"{ip.vendor}__{ip.name}__{so_lib}.{sim_str}.so"
        so_lib_path = f"{ip.path}/{ip.scripts_path}/{so_lib}.{sim_str}.so"
        so_libs[so_lib_flat_name] = so_lib_path
    return so_libs


def convert_compilation_args(sim_job):
    args = []
    return args


def convert_elaboration_args(sim_job):
    args = []
    if sim_job.simulator == common.simulators_enum.VIVADO:
        if sim_job.waves or sim_job.cov or sim_job.gui:
            args.append("--debug all")
    return args


def invoke_fsoc(ip, core, sim_job):
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    fsc_args = argparse.Namespace()
    fsc_args.setup       = True
    fsc_args.build       = False
    fsc_args.run         = False
    fsc_args.no_export   = True
    fsc_args.tool        = "xsim"
    fsc_args.target      = ip.dut_fsoc_target
    fsc_args.system      = ip.dut_fsoc_full_name
    fsc_args.backendargs = ""
    fsc_args.system_name = ""
    fsc_args.build_root  = cfg.fsoc_dir + "/" + core.sname
    fsc_args.flag = []
    fsc_cores_root = [core.dir]
    
    
    try:
        fusesoc.main.init_logging(False, False)
        fsc_cfg = fusesoc.main.Config()
        fsc_cm  = fusesoc.main.init_coremanager(fsc_cfg, fsc_cores_root)
        fusesoc.main.run(fsc_cm, fsc_args)
        file_path_partial_name = re.sub(r':', '_', core.name)
        eda_file_dir = cfg.fsoc_dir + "/" + core.sname + "/sim-xsim"
        eda_file_path = eda_file_dir + "/" + file_path_partial_name + "_0.eda.yml"
        core_rel_path = os.path.relpath(core.dir, eda_file_dir)
        if sim_job.simulator == common.simulators_enum.METRICS:
            core_base_path = os.path.relpath(core.dir, cfg.temp_path)
        else:
            core_base_path = "$MIO_" + core.sname.replace("-", "_").upper() + "_SRC_PATH"
        if not os.path.exists(eda_file_path):
            common.fatal("Could not find FuseSoC output file " + eda_file_path)
        else:
            with open(eda_file_path, 'r') as edafile:
                eda_yaml = yaml.load(edafile, Loader=SafeLoader)
                if not eda_yaml:
                    common.fatal("Could not parse FuseSoC output file " + eda_file_path)
                dirs    = []
                files   = []
                defines = []
                for file in eda_yaml['files']:
                    if file['file_type'] == "systemVerilogSource":
                        if 'is_include_file' in file:
                            if 'include_path' in file:
                                dir_path = file['include_path']
                                dir_path = dir_path.replace(core_rel_path, core_base_path)
                                dirs.append(dir_path)
                        else:
                            file_path = file['name']
                            dir_path = pathlib.Path(f"{eda_file_dir}/{file_path}").parent.resolve()
                            dir_path = os.path.relpath(dir_path, eda_file_dir)
                            dir_path = dir_path.replace(core_rel_path, core_base_path)
                            file_path = file_path.replace(core_rel_path, core_base_path)
                            files.append(file_path)
                            dirs.append(dir_path)
                
                for param in eda_yaml['parameters']:
                    if eda_yaml['parameters'][param]['datatype'] == 'bool':
                        new_define = {};
                        #new_define['boolean'] = True
                        new_define['name'] = param
                        if eda_yaml['parameters'][param]['default'] == True:
                            new_define['value'] = 1
                        else:
                            new_define['value'] = 0
                        defines.append(new_define)
                    else:
                        common.fatal("Support for non-bool FuseSoC parameters is not currently implemented")
                sim_nickname = "xsim"
                #if sim_job.simulator == common.simulators_enum.VIVADO:
                #    sim_nickname = "xsim"
                for option in eda_yaml['tool_options'][sim_nickname]['xelab_options']:
                    if (re.match("--define", option)):
                        new_define = {};
                        matches = re.search("--define\s+(\w+)\s*(?:=\s*(\S+))?", option)
                        if matches:
                            new_define['name'] = matches.group(1)
                            if len(matches.groups()) > 2:
                                new_define['value'] = matches.group(2)
                            else:
                                new_define['boolean'] = True
                            defines.append(new_define)
                #else:
                #    common.fatal("FuseSoC cores not yet supported for " + sim_str)
                flist_template = cfg.templateEnv.get_template(f"{sim_str}.flist.j2")
                outputText = flist_template.render(defines=defines, files=files, dirs=dirs)
                flist_file_path = str(pathlib.Path(cfg.temp_path + "/" + file_path_partial_name + "_0.flist").resolve())
                with open(flist_file_path,'w') as flist_file:
                    flist_file.write(outputText)
                flist_file.close()
                
                if sim_job.simulator == common.simulators_enum.METRICS:
                    flist_file_path = os.path.relpath(flist_file_path, cfg.project_dir)
                
                return flist_file_path
    except Exception as e:
        common.fatal("Failed to convert FuseSoC output data for core '" + core.name + "': "+ str(e))


def encrypt_tree(ip_name, location, app):
    global bar
    tcl_script = ""
    files      = []
    
    for file in glob.iglob(location + '**/**', recursive=True):
        file_path = os.path.join(location, file)
        if (file_path[-2:] == ".v") or (file_path[-3:] == ".vh") or (file_path[-3:] == ".sv") or (file_path[-4:] == ".svh"):  # TODO Add support for VHDL files
            common.dbg(f"Adding '{file_path}' to files to be encrypted")
            files.append(file_path)
    
    if app == "viv":
        if not os.path.exists(cfg.encryption_key_path_vivado):
            common.fatal(f"Could not find vivado encryption key at '{cfg.encryption_key_path_vivado}'")
        
        tcl_script += f"encrypt -key {cfg.encryption_key_path_vivado} -lang ver "  # TODO Add support for VHDL files
        for file in files:
            tcl_script += f"{file} "
        common.dbg(f"TCL Script being passed to vivado for encryption: '\n{tcl_script}'")
        tcl_script_path = cfg.temp_path + "/" + ip_name + ".encrypt.viv.tcl"
        try:
            f = open(tcl_script_path, "w")
            f.write(tcl_script)
            f.close()
        except Exception as e:
            common.fatal(f"Failed to write encryption script to disk: {e}")
        launch_eda_bin(cfg.vivado_home + "/vivado", [f"-mode batch", f" -source {tcl_script_path}"], cfg.temp_path, cfg.dbg)
    elif app == "mtr":
        if not os.path.exists(cfg.encryption_key_path_metrics):
            common.fatal(f"Could not find metrics encryption key at '{cfg.encryption_key_path_metrics}'")
        mtr_key_local_path = f"{cfg.temp_path}/metrics.key"
        mtr_key_rel_path = os.path.relpath(mtr_key_local_path, cfg.temp_path)
        common.copy_file(cfg.encryption_key_path_metrics, mtr_key_local_path)
        num_files = len(files)
        with alive_bar(num_files, bar = 'smooth', stats="{eta} estimated", monitor=True, elapsed=True) as bar:
            for file in files:
                filename = os.path.basename(file)
                bar.text(f"{filename}")
                file_dir_path = file.replace(filename, "")
                file_dir_rel_path = os.path.relpath(file_dir_path, cfg.temp_path)
                
                file_r = open(file,mode='r')
                file_text = file_r.read()
                file_r.close()
                file_w = open(file,mode='w')
                file_w.write("`pragma protect begin\n")
                file_w.write(file_text)
                file_w.write("`pragma protect end")
                file_w.close()
                file_rel_path = os.path.relpath(file, cfg.temp_path)
                args = [file_rel_path, f"-i {mtr_key_rel_path}", f"-o {file_rel_path}.e"]
                arg_list_str = ""
                for arg in args:
                    arg_list_str = arg_list_str + f" {arg}"
                if cfg.dbg:
                    arg_list = [f"dvlencrypt -a '{arg_list_str}'"]
                else:
                    arg_list = [f"dvlencrypt -a '{arg_list_str}'"]
                launch_eda_bin(cfg.metrics_home + "/mdc", arg_list, cfg.temp_path, cfg.dbg)
                launch_eda_bin(cfg.metrics_home + "/mdc", ["download", f"{file_rel_path}.e"], cfg.temp_path, cfg.dbg)
                common.move_file(f"{cfg.temp_path}/_downloaded_{filename}.e", file)
                bar()
    else:
        common.fatal("Only vivado and metrics are currently supported for encryption")


def launch_eda_bin(path, args, wd, output=False, shell=True, dry_run=False):
    global eda_processes
    args_str = ""
    commands = []
    for arg in args:
        args_str = args_str + "  " + arg
    if not dry_run:
        os.chdir(wd)
        common.dbg("Launching " + path + " with arguments '" + args_str + "' from " + wd)
        if output:
            p = subprocess.Popen(path + " " + args_str, shell=shell)
        else:
            p = subprocess.Popen(path + " " + args_str + " > /dev/null 2>&1", shell=shell)
        eda_processes.append(p)
        p.wait()
    rel_wd = os.path.relpath(wd, cfg.project_dir)
    commands.append(f"cd {wd}")
    commands.append(f"{path} {args_str}")
    return commands


def kill_all_processes():
    global eda_processes
    for p in eda_processes:
        p.terminate()
        p.wait()
atexit.register(kill_all_processes)


def scan_cmp_log_file_for_errors(log_file_path, sim_job):
    common.dbg("Scanning compilation log file " + log_file_path + " for errors")
    errors = []
    if sim_job.simulator == common.simulators_enum.VIVADO:
        regexes = vivado_cmp_log_error_regexes
    elif sim_job.simulator == common.simulators_enum.VCS:
        regexes = vcs_cmp_log_error_regexes
    elif sim_job.simulator == common.simulators_enum.METRICS:
        regexes = metrics_cmp_log_error_regexes
    elif sim_job.simulator == common.simulators_enum.XCELIUM:
        regexes = xcelium_cmp_log_error_regexes
    elif sim_job.simulator == common.simulators_enum.QUESTA:
        regexes = questa_cmp_log_error_regexes
    elif sim_job.simulator == common.simulators_enum.RIVIERA:
        regexes = riviera_cmp_log_error_regexes
    try:
        for i, line in enumerate(open(log_file_path)):
            for regex in regexes:
                matches = re.search(regex, line)
                if matches:
                    errors.append(line.replace("\n", ""))
    except Exception as e:
        common.fatal("Failed while parsing compilation log file " + log_file_path + ": " + str(e))
    return errors


def scan_elab_log_file_for_errors(log_file_path, sim_job):
    common.dbg("Scanning elaboration log file " + log_file_path + " for errors")
    errors = []
    if sim_job.simulator == common.simulators_enum.VIVADO:
        regexes = vivado_elab_log_error_regexes
    elif sim_job.simulator == common.simulators_enum.VCS:
        regexes = vcs_elab_log_error_regexes
    elif sim_job.simulator == common.simulators_enum.METRICS:
        regexes = metrics_elab_log_error_regexes
    elif sim_job.simulator == common.simulators_enum.XCELIUM:
        regexes = xcelium_elab_log_error_regexes
    elif sim_job.simulator == common.simulators_enum.QUESTA:
        regexes = questa_elab_log_error_regexes
    elif sim_job.simulator == common.simulators_enum.RIVIERA:
        regexes = riviera_elab_log_error_regexes
    try:
        for i, line in enumerate(open(log_file_path)):
            for regex in regexes:
                matches = re.search(regex, line)
                if matches:
                    errors.append(line.replace("\n", ""))
    except Exception as e:
        common.fatal("Failed while parsing elaboration log file " + log_file_path + ": " + str(e))
    return errors


def log_cmp_history_fsoc(core, log_path, sim_job, timestamp_start, timestamp_end):
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    common.dbg("Updating history with FuseSoC core '" + core.name + "' compilation")
    if core.name not in cfg.job_history:
        cfg.job_history[core.name] = {}
    if 'compilation' not in cfg.job_history[core.name]:
        cfg.job_history[core.name]['compilation'] = []
    cfg.job_history[core.name]['compilation'].append({
        "simulator"       : sim_str,
        'timestamp_start' : timestamp_start,
        'timestamp_end'   : timestamp_end,
        'log_path'        : log_path
    })


def log_cmp_history_ip(ip, log_path, sim_job, timestamp_start, timestamp_end):
    ip_str = f"{ip.vendor}/{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    common.dbg("Updating history with IP '" + ip_str + "' compilation")
    if ip_str not in cfg.job_history:
        cfg.job_history[ip_str] = {}
    if 'compilation' not in cfg.job_history[ip_str]:
        cfg.job_history[ip_str]['compilation'] = []
    cfg.job_history[ip_str]['compilation'].append({
        "simulator"       : sim_str,
        'timestamp_start' : timestamp_start,
        'timestamp_end'   : timestamp_end,
        'log_path'        : log_path
    })


def log_cmp_history_vivado_project(ip, log_path_vlog, log_path_vhdl, sim_job, timestamp_start, timestamp_end):
    ip_str = f"{ip.vendor}/{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    common.dbg("Updating history with Vivado Project IP '" + ip_str + "' compilation")
    if ip_str not in cfg.job_history:
        cfg.job_history[ip_str] = {}
    if 'compilation' not in cfg.job_history[ip_str]:
        cfg.job_history[ip_str]['compilation'] = []
    cfg.job_history[ip_str]['compilation'].append({
        "simulator"       : sim_str,
        'timestamp_start' : timestamp_start,
        'timestamp_end'   : timestamp_end,
        'log_path_vlog'   : log_path_vlog,
        'log_path_vhdl'   : log_path_vhdl
    })


def log_elab_history(ip, log_path, sim_job, timestamp_start, timestamp_end):
    ip_str = f"{ip.vendor}/{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    common.dbg("Updating history with IP '" + ip_str + "' elaboration")
    if ip_str not in cfg.job_history:
        cfg.job_history[ip_str] = {}
    if 'elaboration' not in cfg.job_history[ip_str]:
        cfg.job_history[ip_str]['elaboration'] = []
    cfg.job_history[ip_str]['elaboration'].append({
        "simulator"            : sim_str,
        'timestamp_start'      : timestamp_start,
        'timestamp_end'        : timestamp_end,
        'log_path'             : log_path,
        "is_regression"        : sim_job.is_regression,
        "regression_name"      : sim_job.regression_name,
        "regression_timestamp" : sim_job.regression_timestamp
    })


def log_sim_start_history(ip, sim_job, timestamp):
    ip_str = f"{ip.vendor}/{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    common.dbg("Updating history with IP '" + ip_str + "' simulation start")
    if ip_str not in cfg.job_history:
        cfg.job_history[ip_str] = {}
    if 'simulation' not in cfg.job_history[ip_str]:
        cfg.job_history[ip_str]['simulation'] = []
    cfg.job_history[ip_str]['simulation'].append({
        "type"                 : "start",
        "simulator"            : sim_str,
        'timestamp'            : timestamp,
        'test_name'            : sim_job.test,
        'seed'                 : sim_job.seed,
        'waves'                : sim_job.waves,
        'cov'                  : sim_job.cov,
        'gui'                  : sim_job.gui,
        "args"                 : plus_args_to_str(sim_job),
        "is_regression"        : sim_job.is_regression,
        "regression_name"      : sim_job.regression_name,
        "regression_timestamp" : sim_job.regression_timestamp
    })


def log_sim_end_history(ip, sim_job, timestamp_start, timestamp_end):
    ip_str = f"{ip.vendor}/{ip.name}"
    sim_str = common.get_simulator_short_name(sim_job.simulator)
    common.dbg("Updating history with IP '" + ip_str + "' simulation end")
    if ip_str not in cfg.job_history:
        cfg.job_history[ip_str] = {}
    if 'simulation' not in cfg.job_history[ip_str]:
        cfg.job_history[ip_str]['simulation'] = []
    common.dbg(f"{str(len(cfg.job_history[ip_str]['simulation']))} job history items before append()")
    entry = {
        "type"                 : "end",
        "simulator"            : sim_str,
        'timestamp_start'      : timestamp_start,
        'timestamp_end'        : timestamp_end,
        'log_path'             : sim_job.sim_log_file_path,
        'test_name'            : sim_job.test,
        'seed'                 : sim_job.seed,
        'waves'                : sim_job.waves,
        'cov'                  : sim_job.cov,
        'gui'                  : sim_job.gui,
        'path'                 : sim_job.results_path,
        "args"                 : plus_args_to_str(sim_job),
        "is_regression"        : sim_job.is_regression,
        "regression_name"      : sim_job.regression_name,
        "regression_timestamp" : sim_job.regression_timestamp
    }
    cfg.job_history[ip_str]['simulation'].append(entry)
    common.dbg(f"{str(len(cfg.job_history[ip_str]['simulation']))} job history items after append()")


def plus_args_to_str(sim_job):
    cli_args = sim_job.sim_args
    str = ""
    for arg in cli_args:
        str += f" {arg}"
    return str


def plus_args_list_to_str_list(plus_args):
    str_list = []
    for arg in plus_args:
        if plus_args[arg] == "":
            str_list.append(f"+{arg}")
        else:
            str_list.append(f"+{arg}={plus_args[arg]}")
    return str_list





# Copyright 2022 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################



from mio import cfg
from mio import cache
from mio import user
from mio import eal
import sys
import collections
import os
import yaml
import shutil
import re
import platform
import time
from enum import Enum
from yaml.loader import SafeLoader
from datetime import datetime
import hashlib
from alive_progress import alive_bar
from distutils.dir_util import copy_tree
from distutils.dir_util import remove_tree


class simulators_enum(Enum):
    VIVADO  = "viv"
    METRICS = "mtr"
    VCS     = "vcs"
    XCELIUM = "xcl"
    QUESTA  = "qst"
    RIVIERA = "riv"


def dbg(msg):
    if cfg.dbg:
        print("\033[32m\033[1m[mio-dbg] " + msg + " \033[0m")


def info(msg):
    print("\033[34m\033[1m[mio]\033[0m " + msg)


def warning(msg):
    print("\033[33m\033[1m[mio-warning] " + msg + " \033[0m")


def error(msg):
    print("\033[31m\033[1m[mio-error] " + msg + " \033[0m")


def fatal(msg, dump_cache=True):
    print("\033[31m\033[1m[mio-fatal] " + msg + " \033[0m")
    if dump_cache:
        cache.write_caches_to_disk()
        user.write_user_data_to_disk()
    print()
    sys.exit(0)


def banner(msg):
    print("\033[1;35m[mio] *** " + msg + " ***\033[0m")


def prompt(msg):
    return input("\033[35m\033[1m[mio]\033[0m " + msg + " ")


def exit(dump_cache=True):
    dbg("Exiting gracefully")
    if dump_cache:
        cache.write_caches_to_disk()
        user.write_user_data_to_disk()
    #remove_dir(cfg.temp_path)
    print()
    sys.exit(0)


def get_simulator_short_name(simulator):
    if simulator == simulators_enum.VIVADO:
        return "viv"
    elif simulator == simulators_enum.VCS:
        return "vcs"
    elif simulator == simulators_enum.METRICS:
        return "mtr"
    elif simulator == simulators_enum.XCELIUM:
        return "xcl"
    elif simulator == simulators_enum.QUESTA:
        return "qst"
    elif simulator == simulators_enum.RIVIERA:
        return "riv"


def get_simulator_long_name(simulator):
    if simulator == simulators_enum.VIVADO:
        return "vivado"
    elif simulator == simulators_enum.VCS:
        return "vcs"
    elif simulator == simulators_enum.METRICS:
        return "metrics"
    elif simulator == simulators_enum.XCELIUM:
        return "xcelium"
    elif simulator == simulators_enum.QUESTA:
        return "questa"
    elif simulator == simulators_enum.RIVIERA:
        return "riviera"


def merge_dict(d1, d2):
    """
    Modifies d1 in-place to contain values from d2.  If any value
    in d1 is a dictionary (or dict-like), *and* the corresponding
    value in d2 is also a dictionary, then merge them in-place.
    """
    for k, v2 in d2.items():
        v1 = d1.get(k)  # returns None if v1 has no value for this key
        if (type(v1) is dict) and (type(v2) is dict):
            merge_dict(v1, v2)
        else:
            d1[k] = v2


def timestamp():
    return datetime.now().strftime("%Y/%m/%d-%H:%M:%S")


def parse_timestamp(string):
    return datetime.strptime(string, "%Y/%m/%d-%H:%M:%S")


def dir_timestamp(path):
    timestamps = []
    for root, subdirs, files in os.walk(path):
        for file in files:
            file_p = os.path.join(root, file)
            if platform.system() == 'Windows':
                modTimesinceEpoc = os.path.getctime(file_p)
            else:
                statbuf = os.stat(file_p)
                modTimesinceEpoc = statbuf.st_mtime
            modificationTime = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(modTimesinceEpoc))
            timestamps.append(modificationTime)
    if len(timestamps) > 0:
        latest_timestamp = max(timestamps)
        latest_timestamp_str = datetime.strptime(latest_timestamp, '%Y-%m-%d %H:%M:%S')
    else:
        latest_timestamp_str = timestamp()
    return latest_timestamp_str


def create_dir(path):
    try:
        if not os.path.exists(path):
            dbg("Creating directory at " + path)
            os.mkdir(path)
            dbg("Created directory at " + path)
        else:
            dbg("Not creating directory at " + path + " because it already exists")
    except:
        fatal(f"Failed to create directory {path}")


def create_file(path):
    try:
        if not os.path.exists(path):
            dbg("Creating file at " + path)
            f = open(path, "x")
            f.close()
            dbg("Created file at " + path)
        else:
            dbg("Not creating file at " + path + " because it already exists")
    except:
        fatal(f"Failed to create file {path}")


def remove_dir(path):
    if os.path.exists(path):
        dbg(f"Removing directory '{path}'")
        remove_tree(path)


def remove_file(path):
    if os.path.exists(path):
        dbg(f"Removing file '{path}'")
        os.remove(path)


def copy_file(src, dst):
    dbg(f"Copying file from '{src}' to '{dst}'")
    shutil.copyfile(src, dst)


def move_file(src, dst):
    dbg(f"Moving file from '{src}' to '{dst}'")
    shutil.move(src, dst)


def copy_directory(src, dst, symlinks=False, ignore=None):
    dbg(f"Copying directory from '{src}' to '{dst}'")
    try:
        copy_tree(src, dst)
    except Exception as e:
        fatal(f"Failed to copy from '{src}' to '{dst}': {e}")


def calc_file_hash(path):
    hash = hashlib.md5(open(path,'rb').read()).hexdigest()
    dbg("Hash for file " + path  + " is '" + hash + "'")
    return hash


def create_common_files():
    dbg("Creating common directories and files")
    try:
        create_dir(cfg.mio_data_dir        )
        create_dir(cfg.temp_path           )
        create_dir(cfg.dependencies_path   )
        create_dir(cfg.mio_user_dir        )
        create_dir(cfg.user_global_ips_path)
        create_dir(cfg.sim_output_dir      )
        create_file(cfg.job_history_file_path)
        create_file(cfg.ip_cache_file_path   )
        create_file(cfg.fsoc_cache_file_path )
        create_file(cfg.user_file_path       )
        
        with open(cfg.job_history_file_path, 'r') as yaml_file_read:
            ymlr = yaml.load(yaml_file_read, Loader=SafeLoader)
            if not ymlr:
                ymlr = { "history" : {}}
                dbg("Initializing job history file at " + cfg.job_history_file_path)
                with open(cfg.job_history_file_path, 'w') as yaml_file_write:
                    yaml.dump(ymlr, yaml_file_write)
            cfg.job_history = ymlr['history']
        
        with open(cfg.ip_cache_file_path, 'r') as yaml_file_read:
            ymlr = yaml.load(yaml_file_read, Loader=SafeLoader)
            if not ymlr:
                ymlr = {"ip" : {}}
                dbg("Initializing IP cache file at " + cfg.ip_cache_file_path)
                cfg.fresh_ip_cache = True
                with open(cfg.ip_cache_file_path, 'w') as yaml_file_write:
                    yaml.dump(ymlr, yaml_file_write)
        
        with open(cfg.fsoc_cache_file_path, 'r') as yaml_file_read:
            ymlr = yaml.load(yaml_file_read, Loader=SafeLoader)
            if not ymlr:
                ymlr = {"cores" : {}}
                dbg("Initializing FuseSoC cache file at " + cfg.fsoc_cache_file_path)
                cfg.fresh_soc_cache = True
                with open(cfg.fsoc_cache_file_path, 'w') as yaml_file_write:
                    yaml.dump(ymlr, yaml_file_write)
        
        with open(cfg.user_file_path, 'r') as yaml_file_read:
            ymlr = yaml.load(yaml_file_read, Loader=SafeLoader)
            if not ymlr:
                ymlr = {"user" : {}}
                dbg("Initializing User cache file at " + cfg.user_file_path)
                with open(cfg.user_file_path, 'w') as yaml_file_write:
                    yaml.dump(ymlr, yaml_file_write)
    except Exception as e:
        print(f"\033[31m\033[1m[mio-fatal] Could not create work directories and files: {e}\033[0m")
        sys.exit(0)


def parse_dep(string):
    vendor = ""
    ip = ""
    clean_str = string.lower().strip()
    if "/" in clean_str:
        vendor, ip = clean_str.split("/")
    else:
        ip = clean_str
        vendor = ""
    return [vendor, ip]





# Copyright 2021-2022 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################


from mio import clean
from mio import cov
from mio import dox
from mio import results
from mio import sim
from mio import common

import jinja2
import os
import sys
import re
import pathlib
import toml

cli_args = None
dbg             = True
sim_debug       = False
sim_gui         = False
sim_waves       = False
sim_cov         = False
glb_args = {}
glb_cfg  = {}

#version_str = 'Moore.io Client Command Line Interface 1.2.0'
mio_client_dir    = re.sub("cfg.py", "", os.path.realpath(__file__))
mio_data_src_dir  = str(pathlib.Path(os.path.join(mio_client_dir, "../data")).resolve())
mio_template_dir  = str(pathlib.Path(os.path.join(mio_client_dir, "../templates")).resolve())
vivado_home       = os.getenv("MIO_VIVADO_HOME" , '/tools/vivado/2022.1/Vivado/2022.1/bin/')
metrics_home      = os.getenv("MIO_METRICS_HOME", '/tools/metrics/')
vcs_home          = os.getenv("MIO_VCS_HOME"    , '/tools/vcs/'    )
xcelium_home      = os.getenv("MIO_XCELIUM_HOME", '/tools/xcelium/')
questa_home       = os.getenv("MIO_QUESTA_HOME" , '/tools/questa/' )
riviera_home      = os.getenv("MIO_RIVIERA_HOME", '/tools/riviera/')
uvm_dpi_so        = "uvm_dpi"
user_dir          = os.path.expanduser("~")
mio_user_dir      = user_dir + "/.mio"
user_global_ips_path = mio_user_dir + "/vendors"
global_ips_path = ""
org_name      = ""
org_full_name = ""
default_simulator = common.simulators_enum.VIVADO
uvm_version = ""

pwd                   = ""
temp_path             = ""
project_dir           = ""
project_name          = ""
sim_dir               = ""
sim_results_dir       = ""
regr_results_dir      = ""
mio_data_dir          = ""
fsoc_dir              = ""
sim_output_dir        = ""
dependencies_path     = ""
ip_cache_file_path    = ""
fsoc_cache_file_path  = ""
job_history_file_path = ""
commands_file_path    = ""
user_file_path        = mio_user_dir + "/user.yml"
builtin_ip_path       = ""
user_mio_file         = mio_user_dir + "/mio.toml"

sim_timescale   = ""
regression_name = ""
test_suite_name = ""
test_results_path_template = ""
encryption_key_path_vivado = ""
encryption_key_path_metrics = ""

templateLoader = jinja2.FileSystemLoader(searchpath=mio_template_dir)
templateEnv    = jinja2.Environment(loader=templateLoader)

fresh_fsoc_cache = False
fresh_ip_cache = False

ip_paths = []
job_history = {}
configuration = {}
ip_local_cache = {}
ip_external_cache = {}
fsoc_cache = None
target_ip = None
target_ip_name = ""
target_ip_is_local = False

project_toml_file_path = "./mio.toml"



def set_pwd(wd):
    global pwd
    pwd = str(wd.resolve())




def set_project_dir(path):
    global project_dir 
    global mio_data_dir
    global temp_path
    global fsoc_dir
    global sim_output_dir
    global dependencies_path
    global builtin_ip_path
    global ip_cache_file_path
    global fsoc_cache_file_path
    global job_history_file_path
    global commands_file_path
    project_dir           = path
    mio_data_dir          = project_dir + "/.mio"
    temp_path             = mio_data_dir + '/temp'
    fsoc_dir              = mio_data_dir + "/fsoc"
    sim_output_dir        = mio_data_dir + "/sim"
    dependencies_path     = mio_data_dir + "/vendors"
    builtin_ip_path       = mio_data_src_dir + "/ip"
    commands_file_path    = mio_data_dir + "/commands.yml"
    fsoc_cache_file_path  = mio_data_dir + "/fsoc_cache.yml"
    ip_cache_file_path    = mio_data_dir + "/ip_cache.yml"
    job_history_file_path = mio_data_dir + "/job_history.yml"




def load_configuration():
    global configuration
    global project_name
    global sim_dir
    global sim_results_dir
    global regr_results_dir
    global ip_paths
    global test_results_path_template
    global encryption_key_path_vivado
    global encryption_key_path_metrics
    global org_name
    global org_full_name
    global global_ips_path
    global default_simulator
    global uvm_version
    global sim_timescale
    
    project_name      = configuration.get("project", {}).get("name")
    #org_name          = user.user_data['org-name']
    #org_full_name     = user.user_data['org-full-name']
    sim_dir           = os.path.join(project_dir, configuration.get("simulation", {}).get("root-path"))
    sim_results_dir   = os.path.join(sim_dir    , configuration.get("simulation", {}).get("results-dir"))
    regr_results_dir  = os.path.join(sim_dir    , configuration.get("simulation", {}).get("regressions-dir"))
    uvm_version                = configuration.get("simulation", {}).get("uvm-version").strip()
    sim_timescale              = configuration.get("simulation", {}).get("timescale").strip()
    test_results_path_template = configuration.get("simulation", {}).get("test-result-path-template").strip()
    default_simulator_str      = configuration.get("simulation", {}).get("default-simulator").strip()
    
    encryption_key_path_vivado  = configuration.get("encryption", {}).get("vivado-key-path" ).strip().replace("~", user_dir)
    encryption_key_path_metrics = configuration.get("encryption", {}).get("metrics-key-path").strip().replace("~", user_dir)
    
    org_name      = configuration.get("org", {}).get("name").strip()
    org_full_name = configuration.get("org", {}).get("full-name").strip()
    
    global_ips_path = configuration.get("ip", {}).get("global-paths")
    ip_paths        = configuration.get("ip", {}).get("paths")
    
    if not encryption_key_path_vivado == None:
        encryption_key_path_vivado = encryption_key_path_vivado.replace("~", user_dir)
    if not encryption_key_path_metrics == None:
        encryption_key_path_metrics = encryption_key_path_metrics.replace("~", user_dir)
    
    if default_simulator_str == "viv":
        default_simulator = common.simulators_enum.VIVADO
    elif default_simulator_str == "vcs":
        default_simulator = common.simulators_enum.VCS
    elif default_simulator_str == "mtr":
        default_simulator = common.simulators_enum.METRICS
    elif default_simulator_str == "xcl":
        default_simulator = common.simulators_enum.XCELIUM
    elif default_simulator_str == "qst":
        default_simulator = common.simulators_enum.QUESTA
    elif default_simulator_str == "riv":
        default_simulator = common.simulators_enum.RIVIERA
    else:
        common.warning(f"Default simulator selected ('{default_simulator_str}') is invalid.  Using vivado.")
        default_simulator = common.simulators_enum.VIVADO




def find_project_descriptor():
    global project_toml_file_path
    found_file = False
    current_dir = pwd
    while found_file == False:
        if not os.path.exists(current_dir):
            return False
        else:
            project_toml_file_path = os.path.join(current_dir, "mio.toml")
            project_toml_file_path = str(pathlib.Path(project_toml_file_path).resolve())
            if not os.path.exists(project_toml_file_path):
                current_dir = os.path.join(current_dir, "..")
            else:
                found_file = True
                set_project_dir(current_dir)
                common.dbg("Found Moore.io project TOML Configuration file at " + project_toml_file_path)
    return found_file




def load_tree():
    global configuration
    builtin_toml_file_path = os.path.join(mio_data_src_dir, "mio.toml")
    user_toml_file_path    = os.path.join(mio_user_dir    , "mio.toml")
    
    try:
        configuration = toml.load(builtin_toml_file_path)
    except Exception as e:
        common.fatal(f"Failed to parse Built-in Configuration ({builtin_toml_file_path}): {e}", False)
    if os.path.exists(user_toml_file_path):
        try:
            common.merge_dict(configuration, toml.load(user_toml_file_path))
        except Except as e:
            common.fatal(f"Failed to parse User Configuration ({user_toml_file_path}): {e}", False)
        common.dbg("Found Moore.io user TOML Configuration file at " + user_toml_file_path)
    try:
        common.merge_dict(configuration, toml.load(project_toml_file_path))
    except Exception as e:
        common.fatal(f"Failed to parse Project Configuration ({project_toml_file_path}): {e}", False)
    common.dbg("Final configuration:\n" + toml.dumps(configuration))



#!/usr/bin/python3
########################################################################################################################
# Copyright 2022 Datum Technology Corporation
# All rights reserved
########################################################################################################################


########################################################################################################################
# IMPORTS
########################################################################################################################
from datetime import date
import os
import sys
import re
import jinja2
from jinja2 import Template
from mio import common
from mio import new
from mio import cfg
#from vsdx import VisioFile


########################################################################################################################
# GLOBALS
########################################################################################################################
dbg = False
uvm_gen_dir = re.sub("new_agent_parallel.py", "", os.path.realpath(__file__)) + ".."
relative_path_to_template = uvm_gen_dir + "/templates/sets/"
out_path = ""
default_license = "All rights reserved."
name_of_copyright_owner = ""
name = ""
full_name = ""


########################################################################################################################
# TEMPLATE DATA
########################################################################################################################
parameters = { }



########################################################################################################################
# METHODS
########################################################################################################################
def combine_dict(d1, d2):
    d3 = {}
    for d in d1:
        d3[d] = d1[d]
    for d in d2:
        d3[d] = d2[d]
    return d3


def pick_out_path():
    global out_path
    global default_copyright_owner
    if len(sys.argv) > 1:
        out_path = sys.argv[1]
        print("Code will be output to " + out_path)
    if len(sys.argv) > 2:
        default_copyright_owner = sys.argv[2].replace('"', "")
        print("Default copyright owner is " + default_copyright_owner)


def prompt_user_values():
    global out_path
    global name
    global full_name
    global name_of_copyright_owner
    global parameters
    channels = []
    dir_1_str = ""
    dir_2_str = ""
    is_duplex = False
    is_be     = False
    
    default_copyright_owner = cfg.org_full_name
    default_vendor          = cfg.org_name
    
    common.info("Moore.io Template Generator: UVM Advanced Parallel Agent - (v1p0)")
    common.info("******************************************************************")
    common.info("The answers to the following questionnaire will be used to generate the code for your new UVM Agent")
    common.info("")
    
    if out_path == "":
        out_path = common.prompt("Please enter the destination path for this new agent (default: '.'):\n").strip()
        if out_path == "":
            out_path = "."
    
    name_of_copyright_owner = common.prompt("Please enter the name of the copyright holder or hit RETURN for the default (default is '" + default_copyright_owner + "'):\n").strip()
    if name_of_copyright_owner == "":
        name_of_copyright_owner = default_copyright_owner
    
    name_of_vendor = common.prompt("Please enter a vendor name or hit RETURN for the default (default is '" + default_vendor + "'):\n").strip()
    if name_of_vendor == "":
        name_of_vendor = default_vendor
    
    license = common.prompt("Please enter a usage license or hit RETURN for the default (default is '" + default_license + "'):\n").strip()
    if license == "":
        license = default_license
    
    name = common.prompt("Please enter the package name for this agent (ex: 'axi'); this name will be used for all agent types (ex: 'uvma_axi_agent_c'):\n").lower().strip()
    if name == "":
        sys.exit("ERROR: package name cannot be empty.  Exiting.")
    
    full_name = common.prompt("Please enter the full name for this agent (ex: 'Advanced eXtensible Interface'):\n").strip()
    if full_name == "":
        sys.exit("ERROR: full name cannot be empty.  Exiting.")
    
    is_mm_str = common.prompt("Is this agent memory mapped?  [N/y]").strip().lower()
    if is_mm_str == "":
        is_mm = False
    else:
        if is_mm_str == "n" or is_mm_str == "no":
            is_mm = False
        elif is_mm_str == "y" or is_mm_str == "yes":
            is_mm = True
        else:
            sys.exit("ERROR: please enter 'y' or 'n'")
    
    if is_mm:
        mode_1 = common.prompt("Please enter the first mode for this new agent (default: 'mstr'):\n").strip().lower()
        if mode_1 == "":
            mode_1 = "mstr"
        mode_2 = common.prompt("Please enter the second mode for this new agent (default: 'slv'):\n").strip().lower()
        if mode_2 == "":
            mode_2 = "slv"
        is_be_str = common.prompt("Does this agent support byte-enabled addressing?  [Y/n]").strip().lower()
        if is_be_str == "":
            is_be = True
        else:
            if is_be_str == "n" or is_be_str == "no":
                is_be = False
            elif is_be_str == "y" or is_be_str == "yes":
                is_be = True
            else:
                sys.exit("ERROR: please enter 'y' or 'n'")
        num_segments_str = common.prompt("How many channels are on the physical interface? (Ex: AXI has 5):\n").strip()
        num_segments = int(num_segments_str)
        if num_segments_str == "" or (num_segments <= 0):
            sys.exit("ERROR: please enter an integer larger than 0")
        if num_segments > 5:
            sys.exit("ERROR: please enter an integer smaller or equal to 5")
        if num_segments > 1:
            for ii in range(num_segments):
                channel_name = common.prompt("Please enter the name for channel #" + str(ii+1) + ":\n").strip().lower()
                if channel_name == "":
                    sys.exit("ERROR: channel name cannot be empty.  Exiting")
                tx_name = common.prompt("Please select which agent mode is the 'tx' for channel '" + channel_name + "': '" + mode_1 + "' or '" + mode_2 + "'\n").strip().lower()
                if tx_name != mode_1 and tx_name != mode_2:
                    sys.exit("ERROR: channel tx must be '" + mode_1 + "' or '" + mode_2 + "'.  Exiting")
                if tx_name == mode_1:
                    rx_name = mode_2
                else:
                    rx_name = mode_1
                channels.append({
                    "tx"   : tx_name,
                    "rx"   : rx_name,
                    "name" : channel_name,
                    "index": ii
                })
        else:
            channels.append({
                "tx"   : mode_1,
                "rx"   : mode_2,
                "name" : "phy",
                "index": 0
            })
    else:
        mode_1 = common.prompt("Please enter the first mode for this new agent (default: 'client'):\n").strip().lower()
        if mode_1 == "":
            mode_1 = "client"
        mode_2 = common.prompt("Please enter the second mode for this new agent (default: 'mac'):\n").strip().lower()
        if mode_2 == "":
            mode_2 = "mac"
        
        is_duplex_str = common.prompt("Is this agent duplex?  [Y/n]").strip().lower()
        if is_duplex_str == "":
            num_segments = 2
            is_duplex = False
        else:
            if is_duplex_str == "n" or is_duplex_str == "no":
                num_segments = 1
                is_duplex = False
            elif is_duplex_str == "y" or is_duplex_str == "yes":
                num_segments = 2
                is_duplex = True
            else:
                common.fatal("ERROR: please enter 'y' or 'n'")
        if is_duplex:
            dir_1_str = common.prompt("Please enter the first direction for this new agent (default: 'tx'):\n").strip().lower()
            if dir_1_str == "":
                dir_1_str = "tx"
            dir_2_str = common.prompt("Please enter the second direction for this new agent (default: 'rx'):\n").strip().lower()
            if dir_2_str == "":
                dir_2_str = "rx"
            channels.append({
                "tx"   : mode_1,
                "rx"   : mode_2,
                "name" : dir_1_str,
                "index": 0
            })
            channels.append({
                "tx"   : mode_2,
                "rx"   : mode_1,
                "name" : dir_2_str,
                "index": 1
            })
        else:
            channels.append({
                "tx"   : mode_1,
                "rx"   : mode_2,
                "name" : "phy",
                "index": 0
            })
    
    mm_agent_files[svg_agent_block_diagrams [num_segments-1]] = "uvma_{{ name }}/docs/agent_block_diagram.svg"
    mm_agent_files[vsdx_agent_block_diagrams[num_segments-1]] = "uvma_{{ name }}/docs/agent_block_diagram.vsdx"
    if is_duplex:
        stream_agent_files[svg_agent_block_diagrams   [-1]] = "uvma_{{ name }}/docs/agent_block_diagram.svg"
        stream_agent_files[vsdx_agent_block_diagrams  [-1]] = "uvma_{{ name }}/docs/agent_block_diagram.vsdx"
        stream_env_files[svg_stream_env_block_diagrams [1]] = "uvme_{{ name }}_st/docs/env_block_diagram.svg"
        stream_env_files[vsdx_stream_env_block_diagrams[1]] = "uvme_{{ name }}_st/docs/env_block_diagram.vsdx"
    else:
        stream_agent_files[svg_agent_block_diagrams    [0]] = "uvma_{{ name }}/docs/agent_block_diagram.svg"
        stream_agent_files[vsdx_agent_block_diagrams   [0]] = "uvma_{{ name }}/docs/agent_block_diagram.vsdx"
        stream_env_files[svg_stream_env_block_diagrams [0]] = "uvme_{{ name }}_st/docs/env_block_diagram.svg"
        stream_env_files[vsdx_stream_env_block_diagrams[0]] = "uvme_{{ name }}_st/docs/env_block_diagram.vsdx"
    
    parameters = {
        "name"                    : name,
        "full_name"               : full_name,
        "name_of_copyright_owner" : name_of_copyright_owner,
        "license"                 : license,
        "year"                    : date.today().year,
        "mm"                      : is_mm,
        "sbe"                     : is_be,
        "mode_1"                  : mode_1,
        "mode_2"                  : mode_2,
        "dir_1"                   : dir_1_str,
        "dir_2"                   : dir_2_str,
        "duplex"                  : is_duplex,
        "segments"                : channels,
        "num_segments"            : num_segments,
        "multisegment"            : (len(channels) > 1),
        "vendor"                  : name_of_vendor
    }
    
    # Flatten some parameters for svg
    sequencer = 0
    driver = 0
    chan_id = 0
    for channel in channels:
        parameters['sqrn' + str(sequencer)] = mode_1 + "_" + channel['name']
        parameters['sqrt' + str(sequencer)] = name + "_" + mode_1 + "_" + channel['name']
        sequencer += 1
        parameters['sqrn' + str(sequencer)] = mode_2 + "_" + channel['name']
        parameters['sqrt' + str(sequencer)] = name + "_" + mode_2 + "_" + channel['name']
        sequencer += 1
        parameters['drvn' + str(driver)] = mode_1 + "_" + channel['name']
        parameters['drvt' + str(driver)] = name + "_" + mode_1 + "_" + channel['name']
        parameters['dmp' + str(driver)] = "drv_" + mode_1 + "_" + channel['name'] + "_mp"
        driver += 1
        parameters['drvn' + str(driver)] = mode_2 + "_" + channel['name']
        parameters['drvt' + str(driver)] = name + "_" + mode_2 + "_" + channel['name']
        parameters['dmp' + str(driver)] = "drv_" + mode_2 + "_" + channel['name'] + "_mp"
        driver += 1
        parameters['mmp' + str(chan_id)] = "mon_" + channel['name'] + "_mp"
        chan_id += 1


def gen_agent():
    common.fatal("Not implemented yet.")


def print_end_message():
    global out_path
    global name
    global parameters
    common.banner("IPs successfully generated:")
    common.info("  * " + out_path + "/uvma_" + name)
    common.info("  * " + out_path + "/uvme_" + name + "_st")
    common.info("  * " + out_path + "/uvmt_" + name + "_st")


########################################################################################################################
# ENTRY POINT
########################################################################################################################
def main():
    #pick_out_path()
    prompt_user_values()
    gen_agent()
    print_end_message()



# Copyright 2022 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################


from mio import common
from mio import cfg
from mio import clean
from mio import dox
from mio import results
from mio import sim
from mio import eal
from mio import cache

import os
import yaml
from yaml.loader import SafeLoader
from datetime import datetime


#xcrg  -dir a1  -dir b1  -db_name d1  -db_name   d2  -merge_dir    m1   -merge_db_name   n1 -log result.txt  -report_format   html  -report_dir    report1
def gen_cov_report(ip_str, is_regression=False, test_suite="", regression_name="", regression_timestamp=""):
    vendor, name = common.parse_dep(ip_str)
    if vendor == "":
        ip = cache.get_anon_ip(name, True)
    else:
        ip = cache.get_ip(vendor, name, True)
    sim_lib = f"{ip.vendor}/{ip.name}"
    dir_string = ""
    db_name_string = ""
    if is_regression:
        cov_path    = cfg.regr_results_dir + f"/{ip.name}_{regression_name}/{regression_timestamp}/cov"
        merge_path  = cfg.regr_results_dir + f"/{ip.name}_{regression_name}/{regression_timestamp}/cov/merge"
        report_path = cfg.regr_results_dir + f"/{ip.name}_{regression_name}/{regression_timestamp}/cov/report"
    else:
        cov_path    = cfg.sim_dir + f"/cov"
        merge_path  = cfg.sim_dir + f"/cov/merge/{ip.name}/"
        report_path = cfg.sim_dir + f"/cov/reports/{ip.name}/"
    common.create_dir(cov_path)
    common.create_dir(merge_path)
    common.create_dir(report_path)
    
    merge_string = "-merge_dir " + merge_path + " -merge_db_name " + sim_lib
    now = datetime.now()
    timestamp = now.strftime("%Y/%m/%d-%H:%M:%S")
    if not sim_lib in cfg.job_history:
        common.fatal(f"No record of simulations for IP '{sim_lib}'")
    else:
        if not 'simulation' in cfg.job_history[sim_lib]:
            common.fatal(f"No record of simulations for IP '{sim_lib}'")
        else:
            for sim in cfg.job_history[sim_lib]['simulation']:
                if sim['simulator'] != 'vivado':
                    continue
                if sim['type'] != 'end':
                    continue
                cov_path = sim['path'] + "/cov"
                if sim['cov']:
                    dir_string     = dir_string + " -dir " + cov_path
                    db_name_string = db_name_string + " -db_name " + sim['test_name'] + "_" + str(sim['seed'])
    wd = cfg.sim_output_dir + "/viv/cov_wd"
    eal.launch_eda_bin(cfg.vivado_home + "/xcrg", [dir_string, db_name_string, merge_string, "-report_format html", f"-report_dir {report_path}"], wd)
    return report_path




# Copyright 2022 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################


from mio import common
from mio import cfg
from mio import cov
from mio import dox
from mio import results
from mio import sim
from gen import new_agent_serial
from gen import new_agent_parallel
from gen import new_block
from gen import new_lib
from gen import new_singleton
from gen import new_ss
from gen import new_ral

import jinja2
from jinja2 import Template
import zipfile
import shutil
import os
import sys


templates = ["block_tb","ss_tb","ral","parallel_agent","serial_agent","lib","vseq","test","obj","comp","reg_block","reg"]
singleton_names = ["vseq","test","obj","comp","reg_block","reg"]

def menu(template_name=""):
    global singleton_names
    template_num = 0
    prompt = True
    if template_name != "":
        template_num = 0
        for template in templates:
            if template_name == template:
                prompt = False
                break
            template_num += 1
    else:
        prompt = True
    if prompt:
        common.info("***********************************************************************************************************************")
        common.info("  The following Moore.io UVM Framework templates are available:")
        common.info("***********************************************************************************************************************")
        common.info("  [ 0] - UVM Block-Level Environment & Test Bench")
        common.info("  [ 1] - UVM Sub-System Environment & Test Bench")
        common.info("  [ 2] - UVM Register Model (from .csv)")
        common.info("  [ 3] - UVM Agent - Parallel Interface")
        common.info("  [ 4] - UVM Agent - Serial Interface")
        common.info("  [ 5] - UVM Library")
        common.info("-------------------------------------------------")
        common.info("  [ 6] - UVM Virtual Sequence")
        common.info("  [ 7] - UVM Test")
        common.info("  [ 8] - UVM Object")
        common.info("  [ 9] - UVM Component")
        common.info("  [10] - UVM Register Block")
        common.info("  [11] - UVM Register")
        common.warning("Not yet implemented.")
        common.exit()
        template_num_str = common.prompt("Please enter the index of the template you wish to run: ").strip()
        template_num = int(template_num_str)
    
    if template_num == 0:
        new_block.main()
    elif template_num == 1:
        new_ss.main()
    elif template_num == 2:
        new_ral.main()
    elif template_num == 3:
        new_agent_parallel.main()
    elif template_num == 4:
        new_agent_serial.main()
    elif template_num == 5:
        new_lib.main()
    else:
        if template_num >= 6 and template_num <= 11:
            new_singleton.main(singleton_names[template_num-6])
        else:
            common.fatal("Invalid option: " + str(template_num))






# Copyright 2022 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################

from mio import common
from mio import cfg
from mio import cache
from mio import eal

import yaml
from yaml import SafeLoader

import datetime
from datetime import datetime as date
import requests
import getpass
import tarfile
import json
from pathlib import Path
from base64 import b64encode
import os
import sys
import shutil
from alive_progress import alive_bar


base_url      = "https://mooreio.com"
jwt_endpoint  = base_url + "/api/authenticate"
user_data = {}
org_name = ""
org_full_name = ""


def login(username="", password="", force=False):
    global user_data
    ask_username = True
    ask_password = True
    
    if not force:
        if (('username' in user_data) and ('token' in user_data) and ('expiration' in user_data)):
            if user_data['username'] != "":
                ask_username = False
                if user_data['token'] != "":
                    expiration_date = common.parse_timestamp(user_data['expiration'])
                    if expiration_date > date.now():
                        token    = user_data['token']
                        username = user_data['username']
                        common.dbg(f"User credentials and token are still valid.  Using token '{token}'")
                        return token
                    else:
                        common.dbg(f"User credentials are valid but token has expired.")
    
    if force:
        if (username != "") and (username != None):
            ask_username = False
        if (password != "") and (password != None):
            ask_password = False
    
    if ask_username:
        while username == None or username == "":
            username = common.prompt("Please enter your Moore.io account username:")
    else:
        if not force:
            username = user_data['username']
    
    if ask_password:
        while password == None or password == "":
            password = getpass.getpass(prompt='\033[31m\033[1m[mio]\033[0m Password: ')
    
    payload = {
        "username"  : username,
        "password"  : password,
        "rememberMe": "true"
    }
    try:
        user_data['username'] = username
        jwt_token_response = requests.post(jwt_endpoint, json=payload)
        common.dbg(f"JSON Response from Moore.io Authentication: '{str(jwt_token_response)}'")
        user_data['token'] = jwt_token_response.json()['id_token']
        user_data['expiration'] = date.now() + datetime.timedelta(days=29) # Token is valid for 30, but play it safe
        user_data['expiration'] = user_data['expiration'].strftime("%Y/%m/%d-%H:%M:%S")
        common.banner(f"Successfully logged in as '{username}'")
    except Exception as e:
        common.fatal(f"Failed to log in: '{e}'")
    return user_data['token']


def get_token():
    if user_data != None:
        if (('username' in user_data) and ('token' in user_data) and ('expiration' in user_data)):
            if user_data['username'] != "":
                if user_data['token'] != "":
                    expiration_date = common.parse_timestamp(user_data['expiration'])
                    if expiration_date > date.now():
                        token    = user_data['token']
                        username = user_data['username']
                        common.dbg(f"User credentials and token are still valid.  Using token '{token}'")
                        return token
                    else:
                        common.dbg(f"User credentials are valid but token has expired.")
                        return None


def load_user_data():
    global user_data
    global org_name
    global org_full_name
    
    if not os.path.exists(cfg.mio_user_dir):
        first_time_setup()
    else:
        try:
            with open(cfg.user_file_path, 'r') as yamlfile:
                user_data = yaml.load(yamlfile, Loader=SafeLoader)
                if 'org-name' not in user_data:
                    first_time_setup()
                elif 'org-full-name' not in user_data:
                    first_time_setup()
                org_name      = user_data['org-name'].strip()
                org_full_name = user_data['org-full-name'].strip()
        except Exception as e:
            common.warning(f"User data file is damaged/malformed.  Starting fresh.")
            first_time_setup()


def write_user_data_to_disk():
    try:
        with open(cfg.user_file_path, 'w') as yaml_file_write:
            yaml.dump(user_data, yaml_file_write)
    except Exception as e:
        print("\033[31m\033[1m[mio-fatal] Could not write User data to disk \033[0m: " + str(e))
        sys.exit(0)


def first_time_setup():
    global user_data
    global org_name
    global org_full_name
    common.banner("New workstation setup")
    org_name = ""
    org_full_name = ""
    while org_name == "":
        org_name = common.prompt("Please enter your organization's name (aka vendor name) (ex: 'Acme'):")
    while org_full_name == "":
        org_full_name = common.prompt("Please enter your organization's full legal name (ex: 'Acme Enterprises Inc.'):")
    user_data['org-name'] = org_name
    user_data['org-full-name'] = org_full_name
    common.create_dir(cfg.mio_user_dir)
    common.create_file(cfg.user_mio_file)
